@@ -10,9 +10,10 @@ foo.innerHTML = string or document.write(string).
 
 #include "eb.h"
 
-/* The current (foreground) edbrowse window.
- * This is replaced with a stub when run within the javascript process. */
+/* The current (foreground) edbrowse window and frame.
+ * These are replaced with stubs when run within the javascript process. */
 struct ebWindow *cw;
+struct ebFrame *cf;
 
 /* traverse the tree of nodes with a callback function */
 nodeFunction traverse_callback;
@@ -23,9 +24,20 @@ static void jsNode(struct htmlTag *node, bool opentag);
 
 static void processStyles(jsobjtype so, const char *stylestring);
 
+static bool treeOverflow;
+
 static void traverseNode(struct htmlTag *node)
 {
 	struct htmlTag *child;
+
+	if (node->visited) {
+		treeOverflow = true;
+		debugPrint(4, "node revisit %s %d", node->info->name,
+			   node->seqno);
+		return;
+	}
+	node->visited = true;
+
 	(*traverse_callback) (node, true);
 	for (child = node->firstchild; child; child = child->sibling)
 		traverseNode(child);
@@ -37,11 +49,21 @@ void traverseAll(int start)
 	struct htmlTag *t;
 	int i;
 
+	treeOverflow = false;
+
+	for (i = start; i < cw->numTags; ++i) {
+		t = tagList[i];
+		t->visited = false;
+	}
+
 	for (i = start; i < cw->numTags; ++i) {
 		t = tagList[i];
 		if (!t->parent && !t->slash && t->step < 100)
 			traverseNode(t);
 	}
+
+	if (treeOverflow)
+		debugPrint(3, "malformed tree!");
 }				/* traverseAll */
 
 static int nopt;		/* number of options */
@@ -96,6 +118,7 @@ static void makeButton(void)
 	t->controller = currentForm;
 	t->itype = INP_SUBMIT;
 	t->value = emptyString;
+	t->step = 1;
 	linkinTree(currentForm, t);
 }				/* makeButton */
 
@@ -116,6 +139,7 @@ struct htmlTag *findOpenList(struct htmlTag *t)
 }				/* findOpenList */
 
 /*********************************************************************
+tidy workaround functions.
 Consider html like this.
 <body>
 <A href=http://www.edbrowse.org>Link1
@@ -123,9 +147,9 @@ Consider html like this.
 <A href=http://www.edbrowse.org>Link3
 </body>
 Each anchor should close the one before, thus rendering as
-{Link1} {Link2} {Link3}
+ {Link1} {Link2} {Link3}
 But tidy does not do this; it allows anchors to nest, thus
-{Link1{Link2{Link3}}}
+ {Link1{Link2{Link3}}}
 Not a serious problem really, it just looks funny.
 And yes, html like this does appear in the wild.
 This routine restructures the tree to move the inner anchor
@@ -162,24 +186,134 @@ static void nestedAnchors(int start)
 		}
 		a2->sibling = 0;
 
-/* and link a2 up next to a1 */
+/* then link a2 up next to a1 */
 		a2->parent = a1->parent;
 		a2->sibling = a1->sibling;
 		a1->sibling = a2;
 	}
 }				/* nestedAnchors */
 
+/*********************************************************************
+Bad html will derail tidy, so that <a><div>stuff</div></a>
+will push div outside the anchor, to render as  {} stuff
+m.facebook.com is loaded with them.
+Here is a tiny example.
+
+<body>
+<input type=button name=whatever value=hohaa>
+<a href="#bottom"><div>Cognitive business is here</div></a>
+</body>
+
+This routine puts it back.
+An anchor with no children followd by div
+moves div under the anchor.
+For a while I had this function commented out, like it caused a problem,
+but I can't see why or how, so it's back, and facebook looks better.
+
+As an after kludge, don't move <div> under <a> if <div> has an anchor beneath it.
+That could create nested anchors, which we already worked hard to get rid of.   Eeeeeeesh.
+
+This and other tidy workaround functions are based on heuristics,
+and suffer from false positives and false negatives,
+the former being the more serious problem -
+i.e. we rearrange the tree when we shouldn't.
+Even when we do the right thing, there is another problem,
+innerHTML is wrong, and doesn't match the tree of nodes
+or the original source.
+innerHTML comes to us from tidy, after it has fixed (sometimes broken) things.
+Add <script> to the above, browse, jdb, and look at document.body.innerHTML.
+It does not match the source, in fact it represent the tree *before* we fixed it.
+There really isn't anything I can do about that.
+In so many ways, the better approach is to fix tidy, but sometimes that is out of our hands.
+*********************************************************************/
+
+static bool tagBelow(struct htmlTag *t, int action)
+{
+	struct htmlTag *c;
+
+	if (t->action == action)
+		return true;
+	for (c = t->firstchild; c; c = c->sibling)
+		if (tagBelow(c, action))
+			return true;
+	return false;
+}				/* tagBelow */
+
+static void emptyAnchors(int start)
+{
+	int j;
+	struct htmlTag *a0, *div, *up;
+
+	for (j = start; j < cw->numTags; ++j) {
+		a0 = tagList[j];
+		if (a0->action != TAGACT_A || a0->firstchild)
+			continue;
+// anchor no children
+		for (up = a0; up; up = up->parent)
+			if (up->sibling)
+				break;
+		if (!up || !(div = up->sibling) || div->action != TAGACT_DIV)
+			continue;
+// div follows
+/* would moving this create nested anchors? */
+		if (tagBelow(div, TAGACT_A))
+			continue;
+/* shouldn't have inputs or forms in an anchor. */
+		if (tagBelow(div, TAGACT_INPUT))
+			continue;
+		if (tagBelow(div, TAGACT_FORM))
+			continue;
+		up->sibling = div->sibling;
+		a0->firstchild = div;
+		div->parent = a0;
+		div->sibling = 0;
+	}
+}				/* emptyAnchors */
+
+/*********************************************************************
+If a form is in a table, but not in tr or td, it closes immediately,
+and all the following inputs are orphaned.
+Check for an empty form beneath table, and move all the following siblings
+down into the form.
+*********************************************************************/
+
+static void tableForm(int start)
+{
+	int j;
+	struct htmlTag *form, *table, *t;
+
+	for (j = start; j < cw->numTags; ++j) {
+		form = tagList[j];
+		if (form->action != TAGACT_FORM || form->firstchild)
+			continue;
+		t = form;
+		for (table = form->sibling; table; table = table->sibling) {
+			if (table->action == TAGACT_TABLE &&
+			    tagBelow(table, TAGACT_INPUT)) {
+/* table with inputs below; move it to form */
+/* hope this doesn't break anything */
+				table->parent = form;
+				form->firstchild = table;
+				t->sibling = table->sibling;
+				table->sibling = 0;
+				break;
+			}
+			t = table;
+		}
+	}
+}				/* tableForm */
+
 void formControl(struct htmlTag *t, bool namecheck)
 {
 	int itype = t->itype;
 	char *myname = (t->name ? t->name : t->id);
-	struct htmlTag *cf = currentForm;
-	if (!cf) {
+	struct htmlTag *cform = currentForm;
+	if (!cform) {
 /* nodes could be created dynamically, not through html */
-		cf = findOpenTag(t, TAGACT_FORM);
+		cform = findOpenTag(t, TAGACT_FORM);
 	}
-	if (cf)
-		t->controller = cf;
+	if (cform)
+		t->controller = cform;
 	else if (itype != INP_BUTTON)
 		debugPrint(3, "%s is not part of a fill-out form",
 			   t->info->desc);
@@ -195,6 +329,19 @@ static const char *const inp_types[] = {
 	0
 };
 
+/*********************************************************************
+Here are some other input types that should have additional syntax checks
+performed on them, but as far as this version of edbrowse is concerned,
+they are equivalent to text. Just here to suppress warnings.
+List taken from https://www.tutorialspoint.com/html/html_input_tag.htm
+*********************************************************************/
+
+static const char *const inp_others[] = {
+	"date", "datetime", "datetime-local",
+	"month", "week", "time",
+	"email", "range", "search", "tel", "url", 0
+};
+
 /* helper function for input tag */
 void htmlInputHelper(struct htmlTag *t)
 {
@@ -202,14 +349,16 @@ void htmlInputHelper(struct htmlTag *t)
 	int len;
 	char *myname = (t->name ? t->name : t->id);
 	const char *s = attribVal(t, "type");
-	if (s) {
+	if (stringEqual(t->info->name, "button")) {
+		n = INP_BUTTON;
+	} else if (s) {
 		n = stringInListCI(inp_types, s);
 		if (n < 0) {
-			debugPrint(3, "unrecognized input type %s", s);
+			n = stringInListCI(inp_others, s);
+			if (n < 0)
+				debugPrint(3, "unrecognized input type %s", s);
 			n = INP_TEXT;
 		}
-	} else if (stringEqual(t->info->name, "BUTTON")) {
-		n = INP_BUTTON;
 	}
 	t->itype = n;
 
@@ -283,10 +432,11 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 	int j;
 	int action = t->action;
 	const char *a;		/* usually an attribute */
+	struct htmlTag *u;
 
-#if 0
-	printf("prend %c%s\n", (opentag ? ' ' : '/'), t->info->name);
-#endif
+	debugPrint(6, "prend %c%s %d%s",
+		   (opentag ? ' ' : '/'), t->info->name,
+		   t->seqno, (t->step >= 1 ? "-" : ""));
 
 	if (t->step >= 1)
 		return;
@@ -535,12 +685,27 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 		}
 		break;
 
+	case TAGACT_FRAME:
+		if (opentag)
+			break;
+/* since this browser supports frames, we don't show the text inside */
+		for (u = t->firstchild; u; u = u->sibling) {
+			u->parent = 0;
+			u->deleted = true;
+			u->step = 100;
+		}
+		t->firstchild = 0;
+		break;
+
 	}			/* switch */
 }				/* prerenderNode */
 
 void prerender(int start)
 {
+/* some cleanup routines to rearrange the tree, working around some tidy5 bugs. */
 	nestedAnchors(start);
+	emptyAnchors(start);
+	tableForm(start);
 
 	currentForm = currentSel = currentOpt = NULL;
 	currentTitle = currentScript = currentTA = NULL;
@@ -630,6 +795,9 @@ static jsobjtype establish_js_option(jsobjtype obj, int idx)
 	fo = get_property_object(obj, "form");
 	if (fo)
 		set_property_object(oo, "form", fo);
+	instantiate_array(oo, "childNodes");
+	instantiate_array(oo, "attributes");
+	instantiate(oo, "style", 0);
 
 	return oo;
 }				/* establish_js_option */
@@ -647,8 +815,6 @@ static void establish_inner(jsobjtype obj, const char *start, const char *end,
 	set_property_string(obj, name, s);
 	if (start && end)
 		nzFree((char *)s);
-/* Anything with an innerHTML might also have a style. */
-	instantiate(obj, "style", 0);
 }				/* establish_inner */
 
 static void domLink(struct htmlTag *t, const char *classname,	/* instantiate this class */
@@ -781,7 +947,14 @@ call out to process those and add them to the object */
  processStyles can rely on obj.style existing */
 			if (stylestring)
 				processStyles(so, stylestring);
-/* done with styles - on to forms */
+
+/* Other attributes that are expected by pages, even if they
+ * aren't populated at domLink-time */
+			set_property_string(io, "className", "");
+			set_property_string(io, "class", "");
+			set_property_string(io, "nodeValue", "");
+			instantiate_array(io, "attributes");
+			set_property_object(io, "ownerDocument", cf->docobj);
 
 /* in the special case of form, also need an array of elements */
 			if (stringEqual(classname, "Form"))
@@ -790,7 +963,7 @@ call out to process those and add them to the object */
 
 		if (membername == symname) {
 /* link to document.all */
-			master = get_property_object(cw->docobj, "all");
+			master = get_property_object(cf->docobj, "all");
 			if (master == NULL)
 				return;
 			set_property_object(master, symname, io);
@@ -836,7 +1009,7 @@ call out to process those and add them to the object */
  * a form field named "id". */
 		if (!stringEqual(classname, "Form"))
 			set_property_string(io, "id", idname);
-		master = get_property_object(cw->docobj, "idMaster");
+		master = get_property_object(cf->docobj, "idMaster");
 		set_property_object(master, idname, io);
 	}
 
@@ -850,6 +1023,7 @@ call out to process those and add them to the object */
 
 	if (htmlclass) {
 		set_property_string(io, "className", htmlclass);
+		set_property_string(io, "class", htmlclass);
 	}
 
 	t->jv = io;
@@ -860,7 +1034,7 @@ call out to process those and add them to the object */
 
 	if (stringEqual(classname, "Body")) {
 /* here are a few attributes that come in with the body */
-		set_property_object(cw->docobj, "body", io);
+		set_property_object(cf->docobj, "body", io);
 		set_property_number(io, "clientHeight", 768);
 		set_property_number(io, "clientWidth", 1024);
 		set_property_number(io, "offsetHeight", 768);
@@ -869,11 +1043,11 @@ call out to process those and add them to the object */
 		set_property_number(io, "scrollWidth", 1024);
 		set_property_number(io, "scrollTop", 0);
 		set_property_number(io, "scrollLeft", 0);
-		set_property_object(cw->docobj, "documentElement", io);
+		set_property_object(cf->docobj, "documentElement", io);
 	}
 
 	if (stringEqual(classname, "Head")) {
-		set_property_object(cw->docobj, "head", io);
+		set_property_object(cf->docobj, "head", io);
 	}
 
 }				/* domLink */
@@ -895,7 +1069,7 @@ static void formControlJS(struct htmlTag *t)
 		domLink(t, "Element", 0, "elements", form->jv,
 			isradio | isselect);
 	else
-		domLink(t, "Element", 0, 0, cw->docobj, isradio | isselect);
+		domLink(t, "Element", 0, 0, cf->docobj, isradio | isselect);
 	if (!t->jv)
 		return;
 
@@ -970,77 +1144,121 @@ static void jsNode(struct htmlTag *t, bool opentag)
 		return;
 	t->step = 2;
 
-#if 0
-	printf("decorate %s\n", t->info->name);
-#endif
+/*********************************************************************
+If js is off, and you don't decorate this tree,
+then js is turned on later, and you parse and decorate a frame,
+it might also decorate this tree in the wrong context.
+Needless to say that's not good!
+*********************************************************************/
+	if (t->f0 != cf)
+		return;
+
+	debugPrint(6, "decorate %s %d", t->info->name, t->seqno);
 
 	switch (action) {
+		jsobjtype cd;	/* contentDocument */
+		jsobjtype cdbody;	/* contentDocument.body */
+
 	case TAGACT_TEXT:
-		t->jv = instantiate(cw->docobj, fakePropName(), "TextNode");
+		t->jv = instantiate(cf->docobj, fakePropName(), "TextNode");
 		if (t->jv) {
 			const char *w = t->textval;
 			if (!w)
 				w = emptyString;
 			set_property_string(t->jv, "data", w);
 			set_property_string(t->jv, "nodeName", "text");
+			set_property_number(t->jv, "nodeType", 3);
+/* A text node chould never have children, and does not need childNodes array,
+ * but there is improper html out there <text> <stuff>
+ * which has to put stuff under the text node, so against this
+ * unlikely occurence, I have to create the array. */
+			instantiate_array(t->jv, "childNodes");
 		}
 		break;
 
+	case TAGACT_META:
+		domLink(t, "Meta", 0, "metas", cf->docobj, 0);
+		a = attribVal(t, "content");
+		set_property_string(t->jv, "content", a);
+		set_property_number(t->jv, "nodeType", 1);
+		break;
+
 	case TAGACT_SCRIPT:
-		domLink(t, "Script", "src", "scripts", cw->docobj, 0);
+		domLink(t, "Script", "src", "scripts", cf->docobj, 0);
 		a = attribVal(t, "type");
 		if (a)
 			set_property_string(t->jv, "type", a);
 		a = attribVal(t, "language");
 		if (a)
 			set_property_string(t->jv, "language", a);
+		a = attribVal(t, "src");
+		if (a) {
+			set_property_string(t->jv, "src", a);
+		} else {
+			set_property_string(t->jv, "src", "");
+		}
+		a = attribVal(t, "data");
+		if (a) {
+			set_property_string(t->jv, "data", a);
+		} else {
+			set_property_string(t->jv, "data", "");
+		}
+		set_property_number(t->jv, "nodeType", 1);
 		break;
-
 	case TAGACT_FORM:
-		domLink(t, "Form", "action", "forms", cw->docobj, 0);
+		domLink(t, "Form", "action", "forms", cf->docobj, 0);
 		set_onhandlers(t);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_INPUT:
 		formControlJS(t);
 		if (t->itype == INP_TA)
 			establish_inner(t->jv, t->value, 0, true);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_OPTION:
 		optionJS(t);
 // The parent child relationship has already been established,
 // don't break, just return;
+		set_property_number(t->jv, "nodeType", 1);
 		return;
 
 	case TAGACT_A:
-		domLink(t, "Anchor", "href", "anchors", cw->docobj, 0);
+		domLink(t, "Anchor", "href", "anchors", cf->docobj, 0);
 		set_onhandlers(t);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_HEAD:
-		domLink(t, "Head", 0, "heads", cw->docobj, 0);
+		domLink(t, "Head", 0, "heads", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_BODY:
-		domLink(t, "Body", 0, "bodies", cw->docobj, 0);
+		domLink(t, "Body", 0, "bodies", cf->docobj, 0);
 		set_onhandlers(t);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_OL:
 	case TAGACT_UL:
 	case TAGACT_DL:
-		domLink(t, "Lister", 0, 0, cw->docobj, 0);
+		domLink(t, "Lister", 0, 0, cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_LI:
-		domLink(t, "Listitem", 0, 0, cw->docobj, 0);
+		domLink(t, "Listitem", 0, 0, cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_TABLE:
-		domLink(t, "Table", 0, "tables", cw->docobj, 0);
+		domLink(t, "Table", 0, "tables", cf->docobj, 0);
 /* create the array of rows under the table */
 		instantiate_array(t->jv, "rows");
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_TR:
@@ -1048,75 +1266,93 @@ static void jsNode(struct htmlTag *t, bool opentag)
 			domLink(t, "Trow", 0, "rows", above->jv, 0);
 			instantiate_array(t->jv, "cells");
 		}
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_TD:
 		if ((above = t->controller) && above->jv) {
 			domLink(t, "Cell", 0, "cells", above->jv, 0);
-			establish_inner(t->jv, t->innerHTML, 0, false);
 		}
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_DIV:
-		domLink(t, "Div", 0, "divs", cw->docobj, 0);
-		establish_inner(t->jv, t->innerHTML, 0, false);
+		domLink(t, "Div", 0, "divs", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_OBJECT:
-		domLink(t, "HtmlObj", 0, "htmlobjs", cw->docobj, 0);
-		establish_inner(t->jv, t->innerHTML, 0, false);
+		domLink(t, "HtmlObj", 0, "htmlobjs", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_SPAN:
 	case TAGACT_SUB:
 	case TAGACT_SUP:
 	case TAGACT_OVB:
-		domLink(t, "Span", 0, "spans", cw->docobj, 0);
-		establish_inner(t->jv, t->innerHTML, 0, false);
+		domLink(t, "Span", 0, "spans", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_AREA:
-		domLink(t, "Area", "href", "areas", cw->docobj, 0);
+		domLink(t, "Area", "href", "areas", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_FRAME:
-		domLink(t, "Frame", "src", "frames", cw->winobj, 0);
+		domLink(t, "Frame", "src", "frames", cf->winobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
+/* frames have contentDocument below, even though it is not populated
+ * with the frame's dom as of this version of edbrowse. */
+		cd = instantiate(t->jv, "contentDocument", "Document");
+		cdbody = instantiate(cd, "body", "Body");
+		instantiate(cdbody, "style", 0);
+		instantiate_url(cd, "location", t->href);
+/* perhaps other stuff that a frame document always needs */
 		break;
 
 	case TAGACT_IMAGE:
-		domLink(t, "Image", "src", "images", cw->docobj, 0);
+		domLink(t, "Image", "src", "images", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_P:
-		domLink(t, "P", 0, "paragraphs", cw->docobj, 0);
-		establish_inner(t->jv, t->innerHTML, 0, false);
+		domLink(t, "P", 0, "paragraphs", cf->docobj, 0);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_TITLE:
 		if (cw->ft)
-			set_property_string(cw->docobj, "title", cw->ft);
+			set_property_string(cf->docobj, "title", cw->ft);
+		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	}			/* switch */
 
-/* js tree mirrors the dom tree. */
-	if (t->jv && t->parent && t->parent->jv)
-		run_function_onearg(t->parent->jv, "apch$", t->jv);
+	if (!t->jv)
+		return;		/* nothing else to do */
 
-	if (t->jv && !t->parent) {
+/* js tree mirrors the dom tree. */
+	if (t->parent && t->parent->jv)
+		run_function_onearg(t->parent->jv, "apch1$", t->jv);
+
+	if (!t->parent) {
 		if (innerParent)
-			run_function_onearg(innerParent, "apch$", t->jv);
+			run_function_onearg(innerParent, "apch1$", t->jv);
 /* head and body link to document */
 		else if (action == TAGACT_HEAD || action == TAGACT_BODY)
-			run_function_onearg(cw->docobj, "apch$", t->jv);
+			run_function_onearg(cf->docobj, "apch1$", t->jv);
 	}
 
 /* TextNode linked to document/gc to protect if from garbage collection,
  * but now it is linked to its parent, and even if it isn't,
  * we don't need it hanging around anyways. */
-	if (action == TAGACT_TEXT && t->jv)
-		delete_property(cw->docobj, fakePropLast);
+	if (action == TAGACT_TEXT)
+		delete_property(cf->docobj, fakePropLast);
 
+/* set innerHTML from the source html, if this tag supports it */
+	if (ti->bits & TAG_INNERHTML)
+		establish_inner(t->jv, t->innerHTML, 0, false);
 }				/* jsNode */
 
 /* decorate the tree of nodes with js objects */
@@ -1149,11 +1385,14 @@ static void pushTag(struct htmlTag *t)
 	tagCountCheck();
 }				/* pushTag */
 
+/* first three have to be in this order */
 const struct tagInfo availableTags[] = {
 	{"html", "html", TAGACT_ZERO},
 	{"base", "base reference for relative URLs", TAGACT_BASE, 0, 4},
+	{"object", "an html object", TAGACT_OBJECT, 5, 1},
 	{"a", "an anchor", TAGACT_A, 0, 1},
 	{"input", "an input item", TAGACT_INPUT, 0, 4},
+	{"element", "an input element", TAGACT_INPUT, 0, 4},
 	{"title", "the title", TAGACT_TITLE, 0, 0},
 	{"textarea", "an input text area", TAGACT_TA, 0, 0},
 	{"select", "an option list", TAGACT_SELECT, 0, 0},
@@ -1164,18 +1403,18 @@ const struct tagInfo availableTags[] = {
 	{"font", "a font", TAGACT_NOP, 0, 0},
 	{"center", "centered text", TAGACT_P, 2, 5},
 	{"caption", "a caption", TAGACT_NOP, 5, 0},
-	{"head", "the html header information", TAGACT_HEAD, 0, 4},
-	{"body", "the html body", TAGACT_BODY, 0, 4},
+	{"head", "the html header information", TAGACT_HEAD, 0, 5},
+	{"body", "the html body", TAGACT_BODY, 0, 5},
 	{"text", "a text section", TAGACT_TEXT, 0, 4},
 	{"bgsound", "background music", TAGACT_MUSIC, 0, 4},
 	{"audio", "audio passage", TAGACT_MUSIC, 0, 4},
 	{"meta", "a meta tag", TAGACT_META, 0, 4},
+	{"link", "a link tag", TAGACT_LINK, 0, 4},
 	{"img", "an image", TAGACT_IMAGE, 0, 4},
 	{"image", "an image", TAGACT_IMAGE, 0, 4},
 	{"br", "a line break", TAGACT_BR, 1, 4},
 	{"p", "a paragraph", TAGACT_P, 2, 5},
 	{"div", "a divided section", TAGACT_DIV, 5, 1},
-	{"object", "an html object", TAGACT_OBJECT, 5, 1},
 	{"map", "a map of images", TAGACT_NOP, 5, 0},
 	{"blockquote", "a quoted paragraph", TAGACT_NOP, 10, 1},
 	{"h1", "a level 1 header", TAGACT_NOP, 10, 1},
@@ -1195,12 +1434,12 @@ const struct tagInfo availableTags[] = {
 	{"hr", "a horizontal line", TAGACT_HR, 5, 4},
 	{"form", "a form", TAGACT_FORM, 10, 1},
 	{"button", "a button", TAGACT_INPUT, 0, 4},
-	{"frame", "a frame", TAGACT_FRAME, 2, 4},
-	{"iframe", "a frame", TAGACT_FRAME, 2, 4},
+	{"frame", "a frame", TAGACT_FRAME, 2, 0},
+	{"iframe", "a frame", TAGACT_FRAME, 2, 1},
 	{"map", "an image map", TAGACT_MAP, 2, 4},
 	{"area", "an image map area", TAGACT_AREA, 0, 4},
 	{"table", "a table", TAGACT_TABLE, 10, 1},
-	{"tbody", "a table body", TAGACT_TBODY, 0, 0},
+	{"tbody", "a table body", TAGACT_TBODY, 0, 1},
 	{"tr", "a table row", TAGACT_TR, 5, 1},
 	{"td", "a table entry", TAGACT_TD, 0, 5},
 	{"th", "a table heading", TAGACT_TD, 0, 5},
@@ -1217,7 +1456,6 @@ const struct tagInfo availableTags[] = {
 	{"noframes", "no frames section", TAGACT_NOP, 0, 2},
 	{"embed", "embedded html", TAGACT_MUSIC, 0, 4},
 	{"noembed", "no embed section", TAGACT_NOP, 0, 2},
-	{"object", "an html object", TAGACT_OBJ, 0, 2},
 	{"em", "emphasized text", TAGACT_JS, 0, 0},
 	{"label", "a label", TAGACT_JS, 0, 0},
 	{"strike", "emphasized text", TAGACT_JS, 0, 0},
@@ -1305,12 +1543,18 @@ struct htmlTag *newTag(const char *name)
 	for (ti = availableTags; ti->name[0]; ++ti)
 		if (stringEqualCI(ti->name, name))
 			break;
-	if (!ti->name[0])
-		return 0;
+
+	if (!ti->name[0]) {
+		debugPrint(3, "warning, created node %s reverts to generic",
+			   name);
+		ti = availableTags + 2;
+	}
+
 	if ((action = ti->action) == TAGACT_ZERO)
 		return 0;
 
 	t = (struct htmlTag *)allocZeroMem(sizeof(struct htmlTag));
+	t->f0 = cf;		/* set current frame */
 	t->action = action;
 	t->info = ti;
 	t->seqno = cw->numTags;
@@ -1339,7 +1583,9 @@ void htmlNodesIntoTree(int start, struct htmlTag *attach)
 	treeAttach = attach;
 	tree_pos = start;
 	treeDisable = false;
+	debugPrint(4, "@@tree of nodes");
 	intoTree(0);
+	debugPrint(4, "}\n@@end tree");
 }				/* htmlNodesIntoTree */
 
 /* Convert a list of html nodes, properly nested open close, into a tree.
@@ -1352,16 +1598,22 @@ static void intoTree(struct htmlTag *parent)
 	const char *v;
 	int action;
 
+	if (!parent)
+		debugPrint(4, "root {");
+	else
+		debugPrint(4, "%s %d {", parent->info->name, parent->seqno);
+
 	while (tree_pos < cw->numTags) {
 		t = tagList[tree_pos++];
 		if (t->slash) {
 			if (parent)
 				parent->balance = t, t->balance = parent;
+			debugPrint(4, "}");
 			return;
 		}
 
 		if (treeDisable) {
-			debugPrint(5, "node skip %s", t->info->name);
+			debugPrint(4, "node skip %s", t->info->name);
 			t->step = 100;
 			intoTree(t);
 			continue;
@@ -1374,7 +1626,7 @@ static void intoTree(struct htmlTag *parent)
  * to the children below. */
 			action = t->action;
 			if (action == TAGACT_HEAD) {
-				debugPrint(5, "node skip %s", t->info->name);
+				debugPrint(4, "node skip %s", t->info->name);
 				t->step = 100;
 				treeDisable = true;
 				intoTree(t);
@@ -1382,7 +1634,7 @@ static void intoTree(struct htmlTag *parent)
 				continue;
 			}
 			if (action == TAGACT_BODY) {
-				debugPrint(5, "node pass %s", t->info->name);
+				debugPrint(4, "node pass %s", t->info->name);
 				t->step = 100;
 				intoTree(t);
 				continue;
@@ -1395,7 +1647,7 @@ static void intoTree(struct htmlTag *parent)
 				const char *w = "root";
 				if (treeAttach)
 					w = treeAttach->info->name;
-				debugPrint(5, "node up %s to %s", t->info->name,
+				debugPrint(4, "node up %s to %s", t->info->name,
 					   w);
 				t->parent = treeAttach;
 				if (treeAttach) {
@@ -1414,11 +1666,14 @@ static void intoTree(struct htmlTag *parent)
 		}
 
 /* regular linking through the parent node */
-		t->parent = parent;
+/* Could be treeAttach if this is a frame inside a window */
+		t->parent = (parent ? parent : treeAttach);
 		if (prev) {
 			prev->sibling = t;
 		} else if (parent) {
 			parent->firstchild = t;
+		} else if (treeAttach) {
+			treeAttach->firstchild = t;
 		}
 		prev = t;
 
@@ -1474,13 +1729,13 @@ checkattributes:
 			if (v && !*v)
 				v = 0;
 			if (v) {
-				v = resolveURL(cw->hbase, v);
+				v = resolveURL(cf->hbase, v);
 				cnzFree(t->atvals[j]);
 				t->atvals[j] = v;
-				if (action == TAGACT_BASE && !cw->baseset) {
-					nzFree(cw->hbase);
-					cw->hbase = cloneString(v);
-					cw->baseset = true;
+				if (action == TAGACT_BASE && !cf->baseset) {
+					nzFree(cf->hbase);
+					cf->hbase = cloneString(v);
+					cf->baseset = true;
 				}
 				t->href = cloneString(v);
 			}
@@ -1490,7 +1745,7 @@ checkattributes:
 			if (v && !*v)
 				v = 0;
 			if (v) {
-				v = resolveURL(cw->hbase, v);
+				v = resolveURL(cf->hbase, v);
 				cnzFree(t->atvals[j]);
 				t->atvals[j] = v;
 				if (!t->href)
@@ -1502,7 +1757,7 @@ checkattributes:
 			if (v && !*v)
 				v = 0;
 			if (v) {
-				v = resolveURL(cw->hbase, v);
+				v = resolveURL(cf->hbase, v);
 				cnzFree(t->atvals[j]);
 				t->atvals[j] = v;
 				if (!t->href)
@@ -1523,17 +1778,11 @@ checkattributes:
 	}
 }				/* intoTree */
 
-/* Here is an instance of the edbrowse window that exists for parsing
- * html from inside javascript. It belongs to edbrowse-js,
- * and isn't associated with a particular buffer on the edbrowse side. */
-struct ebWindow in_js_cw;
-
 /* Parse some html, as generated by innerHTML or document.write.
  * This assumes in_js_cw has been set up properly,
  * with the correct winobj and docobj etc. */
 void html_from_setter(jsobjtype inner, const char *h)
 {
-	cw = &in_js_cw;
 	initTagArray();
 	html2nodes(h, false);
 	htmlGenerated = true;
@@ -1565,7 +1814,9 @@ static void processStyles(jsobjtype so, const char *stylestring)
 			skipWhite2(&sv);
 			trimWhite(s);
 			trimWhite(sv);
-			set_property_string(so, s, sv);
+// the property name has to be nonempty
+			if (*s)
+				set_property_string(so, s, sv);
 		}
 	}
 	nzFree(workstring);
