@@ -181,6 +181,7 @@ extern char *userAgents[10], *currentAgent;
 extern char *newlocation;
 extern int newloc_d; /* delay */
 extern bool newloc_r; /* location replaces this page */
+extern struct ebFrame *newloc_f; /* frame calling for new web page */
 extern const char *ebrc_string; /* default ebrc file */
 
 struct eb_curl_callback_data {
@@ -216,6 +217,7 @@ extern int maxproxy;
 struct MIMETYPE {
 	char *type, *desc;
 	char *suffix, *prot, *program;
+	char *urlmatch;
 	char *content;
 	char outtype;
 	bool stream, download;
@@ -265,7 +267,7 @@ extern bool listNA;		/* list nonascii chars */
 extern bool inInput;		/* reading line from standard in */
 extern int fileSize;		/* when reading/writing files */
 extern int mssock;		/* mail server socket */
-extern long hcode;		/* http code, like 404 file not found */
+extern long ht_code;		/* http code, like 404 file not found */
 extern char errorMsg[];		/* generated error message */
 extern char serverLine[];	/* lines to and from the mail server */
 extern int localAccount;	/* this is the smtp server for outgoing mail */
@@ -302,6 +304,9 @@ extern char *home;		/* home directory */
 extern char *recycleBin;	/* holds deleted files */
 extern char *configFile, *sigFile, *sigFileEnd;
 extern char *cookieFile;	/* persistent cookies */
+extern char *cacheDir;	/* directory for a persistent cache of http pages */
+extern int cacheSize; /* in megabytes */
+extern int cacheCount; /* number of cache files */
 
 struct listHead {
 	void *next;
@@ -331,6 +336,7 @@ struct inputChange {
 	int tagno;
 	char major, minor;
 	char filler1, filler2;
+	struct ebFrame *f0;
 	char value[4];
 };
 extern struct listHead inputChangesPending;
@@ -344,6 +350,32 @@ struct lineMap {
 };
 #define LMSIZE sizeof(struct lineMap)
 
+/* an edbrowse frame, as when there are many frames in an html page.
+ * There could be several frames in an edbrowse window or buffer, chained
+ * together in a linked list, but usually there is just one, as when editing
+ * a local file and browsing a simple web page.
+*/
+struct ebFrame {
+	struct ebFrame *next;
+	struct ebWindow *owner;
+	struct htmlTag *frametag;
+	char *fileName;		/* name of file or url */
+	char *firstURL;		/* before http redirection */
+	char *hbase; /* base for href references */
+	bool baseset; /* <base> tag has been seen */
+	bool f_encoded:1; /* filename is url encoded */
+	char *dw;		/* document.write string */
+	int dw_l;		/* length of the above */
+/* The javascript context and window corresponding to this url or frame.
+ * If this is null then javascript is not operational for this frame.
+ * We could still be browsing however, without javascript. */
+	jsobjtype jcx;
+	jsobjtype winobj;
+	jsobjtype docobj;	/* window.document */
+	const struct MIMETYPE *mt;
+};
+extern struct ebFrame *cf;	/* current frame */
+
 /* an edbrowse window */
 struct ebWindow {
 /* windows stack up as you open new files or follow hyperlinks.
@@ -354,31 +386,39 @@ struct ebWindow {
 	int dot, dol;
 /* remember dot and dol for the raw text, when in browse mode */
 	int r_dot, r_dol;
-	char *fileName;		/* name of file or url */
-	char *firstURL;		/* before http redirection */
+	struct ebFrame f0; /* first frame */
+/* is the referrer the original web page or the individual frame? */
 	char *referrer;
 	char *baseDirName;	/* when scanning a directory */
 	char *ft, *fd, *fk;	/* title, description, keywords */
-	char *hbase; /* base for href references */
 	char *mailInfo;
 	char lhs[MAXRE], rhs[MAXRE];	/* remembered substitution strings */
 	struct lineMap *map, *r_map;
 /* The labels that you set with the k command, and access via 'x.
  * Basically, that's 26 line numbers.
- * Number 0 means the label is not set. */
-	int labels[26], r_labels[26];
+ * Number 0 means the label is not set.
+ * But there's one more to mark, when background javascript
+ * adds or deletes lines and we need to keep track of dot. */
+#define MARKLETTERS 27
+#define MARKDOT 26
+	int labels[MARKLETTERS], r_labels[MARKLETTERS];
 /* Next is an array of html tags, generated by the browse command,
  * and used thereafter for hyperlinks, fill-out forms, etc. */
 	struct htmlTag **tags;
 	int numTags, allocTags;
-	const struct MIMETYPE *mt;
+	bool mustrender:1;
+	bool sank:1; /* jSyncup has been run */
 	bool lhs_yes:1;
 	bool rhs_yes:1;
 	bool binMode:1;		/* binary file */
 	bool nlMode:1;		/* newline at the end */
 	bool rnlMode:1;
-/* Two text modes:1; these are incompatible with binMode */
+/* Various text modes, these are incompatible with binMode */
+/* All modes convert to utf8, as that is what pcre understands. */
 	bool utf8Mode:1;
+	bool utf16Mode:1;
+	bool utf32Mode:1;
+	bool bigMode:1; // big-endian
 	bool iso8859Mode:1;
 	bool browseMode:1;	/* browsing html */
 	bool changeMode:1;	/* something has changed in this file */
@@ -386,28 +426,18 @@ struct ebWindow {
 	bool dirMode:1;		/* directory mode */
 	bool undoable:1;	/* undo is possible */
 	bool sqlMode:1;		/* accessing a table */
-	bool baseset:1; /* <base> tag seen */
-	bool f_encoded:1; /* filename is url encoded */
-	char *dw;		/* document.write string */
-	int dw_l;		/* length of the above */
-/* The javascript context and window corresponding to this edbrowse buffer.
- * If this is null then javascript is not operational for this window.
- * We could still be browsing however, without javascript.
- * Refer to browseMode to test for that. */
-	jsobjtype jcx;
-	jsobjtype winobj;
-	jsobjtype docobj;	/* window.document */
 	struct DBTABLE *table;	/* if in sqlMode */
+	time_t nextrender;
 };
 extern struct ebWindow *cw;	/* current window */
-extern struct ebWindow in_js_cw; /* window within edbrowse-js */
+#define foregroundWindow (cw == sessionList[context].lw)
 
 /* quickly grab a tag from the current window via its sequence number:
  * tagList[n] */
 #define tagList (cw->tags)
 
 /* js is running in the current session */
-#define isJSAlive (cw->jcx != NULL && allowJS)
+#define isJSAlive (cf->jcx != NULL && allowJS)
 
 /*********************************************************************
 Temporary cap on the number of lines, so the integer index into cw->map
@@ -453,6 +483,8 @@ struct htmlTag {
 	struct htmlTag *parent, *firstchild, *sibling;
 /* connect <foo> and </foo> */
 	struct htmlTag *balance;
+	struct ebFrame *f0; /* frame that owns this tag */
+	struct ebFrame *f1; /* subordinate frame if this is a <frame> tag */
 	jsobjtype jv;		/* corresponding java variable */
 	int seqno;
 	char *js_file;
@@ -470,6 +502,7 @@ struct htmlTag {
 	bool slash:1;		/* as in </A> */
 	bool textin:1; /* <a> some text </a> */
 	bool deleted:1; /* deleted from the current buffer */
+	bool contracted:1; /* frame is contracted */
 	bool multiple:1;
 	bool rdonly:1;
 	bool clickable:1;	/* but not an input field */
@@ -489,6 +522,7 @@ struct htmlTag {
 	bool onload:1;
 	bool onunload:1;
 	bool doorway:1; /* doorway to javascript */
+	bool visited: 1;
 	char subsup;		/* span turned into sup or sub */
 	uchar itype;		/* input type = */
 	int ninp;		/* number of nonhidden inputs */
@@ -509,7 +543,7 @@ enum {
 	TAGACT_OL, TAGACT_UL, TAGACT_DL,
 	TAGACT_TEXT, TAGACT_BODY, TAGACT_HEAD,
 	TAGACT_MUSIC, TAGACT_IMAGE, TAGACT_BR, TAGACT_IBR, TAGACT_P,
-	TAGACT_BASE, TAGACT_META, TAGACT_PRE, TAGACT_TBODY,
+	TAGACT_BASE, TAGACT_META, TAGACT_LINK, TAGACT_PRE, TAGACT_TBODY,
 	TAGACT_DT, TAGACT_DD, TAGACT_LI, TAGACT_TABLE, TAGACT_TR, TAGACT_TD,
 	TAGACT_DIV, TAGACT_SPAN, TAGACT_HR, TAGACT_OBJECT,
 	TAGACT_FORM, TAGACT_FRAME,
