@@ -20,26 +20,21 @@ static void markAllDead(void)
 {
 	int cx;			/* edbrowse context */
 	struct ebWindow *w;
+	struct ebFrame *f;
 	bool killed = false;
 
 	for (cx = 1; cx < MAXSESSION; ++cx) {
 		w = sessionList[cx].lw;
-		if (!w)
-			continue;
-		if (w->winobj) {
-			w->winobj = 0;
-			w->docobj = 0;
-			w->jcx = 0;
-			killed = true;
-		}
-		while (w != sessionList[cx].fw) {
-			w = w->prev;
-			if (w->winobj) {
-				w->winobj = 0;
-				w->docobj = 0;
-				w->jcx = 0;
+		while (w) {
+			for (f = &w->f0; f; f = f->next) {
+				if (!f->winobj)
+					continue;
+				f->winobj = 0;
+				f->docobj = 0;
+				f->jcx = 0;
 				killed = true;
 			}
+			w = w->prev;
 		}
 	}
 
@@ -227,6 +222,7 @@ static void javaSetsInner(jsobjtype v, const char *newtext, char c)
 	ic->tagno = t->seqno;
 	ic->major = 'i';
 	ic->minor = c;
+	ic->f0 = cf;
 	strcpy(ic->value, newtext);
 	addToListBack(&inputChangesPending, ic);
 }				/* javaSetsInner */
@@ -234,10 +230,10 @@ static void javaSetsInner(jsobjtype v, const char *newtext, char c)
 /* start a document.write */
 void dwStart(void)
 {
-	if (cw->dw)
+	if (cf->dw)
 		return;
-	cw->dw = initString(&cw->dw_l);
-	stringAndString(&cw->dw, &cw->dw_l, "<!DOCTYPE public><body>");
+	cf->dw = initString(&cf->dw_l);
+	stringAndString(&cf->dw, &cf->dw_l, "<!DOCTYPE public><body>");
 }				/* dwStart */
 
 /*********************************************************************
@@ -277,7 +273,7 @@ static void processEffects(void)
 		switch (c) {
 		case 'w':	/* document.write */
 			dwStart();
-			stringAndString(&cw->dw, &cw->dw_l, s);
+			stringAndString(&cf->dw, &cf->dw_l, s);
 			break;
 
 		case 'n':	/* new window */
@@ -308,6 +304,7 @@ static void processEffects(void)
 			ic->tagno = n;
 			ic->major = 't';
 			ic->minor = v[-1];
+			ic->f0 = cf;
 			strcpy(ic->value, s);
 			addToListBack(&inputChangesPending, ic);
 			break;
@@ -315,7 +312,7 @@ static void processEffects(void)
 		case 'c':	/* cookie */
 /* Javascript does some modest syntax checking on the cookie before
  * passing it back to us, so I'm just going to assume it works. */
-			receiveCookie(cw->fileName, s);
+			receiveCookie(cf->fileName, s);
 			break;
 
 		case 'f':
@@ -356,18 +353,24 @@ static void processEffects(void)
  * Returns 0 for ok or -1 for bad read. */
 static int readFromJS(void *data_p, int n)
 {
+	unsigned char *bytes_p = (unsigned char *)data_p;
 	int rc;
 	if (n == 0)
 		return 0;
-	rc = read(pipe_in[0], data_p, n);
-	debugPrint(7, "js read %d", rc);
-	if (rc == n)
-		return 0;
+	while (n > 0) {
+		rc = read(pipe_in[0], bytes_p, n);
+		debugPrint(7, "js read %d", rc);
+		if (rc <= 0) {
 /* Oops - can't read from the process any more */
-	i_puts(MSG_JSEngineRW);
-	js_kill();
-	markAllDead();
-	return -1;
+			i_puts(MSG_JSEngineRW);
+			js_kill();
+			markAllDead();
+			return -1;
+		}
+		n -= rc;
+		bytes_p += rc;
+	}
+	return 0;
 }				/* readFromJS */
 
 static int writeToJS(const void *data_p, int n)
@@ -475,9 +478,20 @@ static int readMessage(void)
 		propval[l] = 0;
 	}
 
+/* sometimes you want to stop at the first js error, but sometimes you don't */
+#if 0
+	if (head.msglen && debugLevel >= 5) {
+		head.highstat = EJ_HIGH_CX_FAIL;
+		head.lowstat = 0;
+		debugPrint(5, "js abort due to error while debugging");
+		puts("edbrowse abort due to error while debugging.");
+		exit(1);
+	}
+#endif
+
 	if (head.highstat == EJ_HIGH_CX_FAIL) {
 		if (head.lowstat == EJ_LOW_VARS)
-			i_puts(MSG_JSEngineVars);
+			i_printf(MSG_JSEngineVars, head.lineno);
 		if (head.lowstat == EJ_LOW_CX)
 			i_puts(MSG_JavaContextError);
 		if (head.lowstat == EJ_LOW_WIN)
@@ -488,7 +502,7 @@ static int readMessage(void)
 			i_puts(MSG_PageDone);
 		else
 			i_puts(MSG_JSSessionFail);
-		freeJavaContext(cw);
+		freeJavaContext(cf);
 /* should I free and zero the property at this point? */
 	}
 
@@ -498,9 +512,9 @@ static int readMessage(void)
 static int writeHeader(void)
 {
 	head.magic = EJ_MAGIC;
-	head.jcx = cw->jcx;
-	head.winobj = cw->winobj;
-	head.docobj = cw->docobj;
+	head.jcx = cf->jcx;
+	head.winobj = cf->winobj;
+	head.docobj = cf->docobj;
 	return writeToJS(&head, sizeof(head));
 }				/* writeHeader */
 
@@ -540,15 +554,17 @@ static void ack5(void)
 }				/* ack5 */
 
 /* Create a js context for the current window.
- * The corresponding js context will be stored in cw->jcx. */
+ * The corresponding js context will be stored in cf->jcx. */
 void createJavaContext(void)
 {
 	if (!allowJS)
 		return;
 
 	if (!js_pid) {
+		int i;
 		js_start();
-		update_var_in_js(1);
+		for (i = 1; i <= 4; ++i)
+			update_var_in_js(i);
 	}
 
 	debugPrint(5, "> create context for session %d", context);
@@ -565,9 +581,9 @@ void createJavaContext(void)
 		return;
 
 /* Copy the context pointer back to edbrowse. */
-	cw->jcx = head.jcx;
-	cw->winobj = head.winobj;
-	cw->docobj = head.docobj;
+	cf->jcx = head.jcx;
+	cf->winobj = head.winobj;
+	cf->docobj = head.docobj;
 
 	setupJavaDom();
 }				/* createJavaContext */
@@ -582,24 +598,24 @@ writeHeader() function above,
 * whibuilds the message assuming cw.
 *********************************************************************/
 
-void freeJavaContext(struct ebWindow *w)
+void freeJavaContext(struct ebFrame *f)
 {
-	if (!w->winobj)
+	if (!f->winobj)
 		return;
 
 	debugPrint(5, "> free context session %d", context);
 
 	head.magic = EJ_MAGIC;
 	head.cmd = EJ_CMD_DESTROY;
-	head.jcx = w->jcx;
-	head.winobj = w->winobj;
-	head.docobj = w->docobj;
+	head.jcx = f->jcx;
+	head.winobj = f->winobj;
+	head.docobj = f->docobj;
 	if (writeToJS(&head, sizeof(head)))
 		return;
 	if (readMessage())
 		return;
 	ack5();
-	w->jcx = w->winobj = 0;
+	f->jcx = f->winobj = 0;
 }				/* freeJavaContext */
 
 void js_shutdown(void)
@@ -637,7 +653,7 @@ char *jsRunScriptResult(jsobjtype obj, const char *str, const char *filename,
 	int rc;
 	char *s;
 
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return 0;
 
 	if (!str || !str[0])
@@ -688,7 +704,7 @@ enum ej_proptype has_property(jsobjtype obj, const char *name)
 {
 	if (whichproc == 'j')
 		return has_property_nat(obj, name);
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return EJ_PROP_NONE;
 
 	debugPrint(5, "> has %s", name);
@@ -712,7 +728,7 @@ void delete_property(jsobjtype obj, const char *name)
 		delete_property_nat(obj, name);
 		return;
 	}
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return;
 
 	debugPrint(5, "> delete %s", name);
@@ -733,7 +749,7 @@ void delete_property(jsobjtype obj, const char *name)
 static int get_property(jsobjtype obj, const char *name)
 {
 	propval = 0;		/* should already be 0 */
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return -1;
 
 	debugPrint(5, "> get %s", name);
@@ -837,7 +853,7 @@ jsobjtype get_property_function(jsobjtype parent, const char *name)
 static int get_array_element(jsobjtype obj, int idx)
 {
 	propval = 0;
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return -1;
 
 	debugPrint(5, "> get [%d]", idx);
@@ -873,7 +889,7 @@ static int set_property(jsobjtype obj, const char *name,
 {
 	int l;
 
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return -1;
 
 	debugPrint(5, "> set %s=%s", name, debugString(value));
@@ -951,7 +967,7 @@ jsobjtype instantiate_array(jsobjtype parent, const char *name)
 	if (whichproc == 'j')
 		return instantiate_array_nat(parent, name);
 
-	if (!allowJS || !cw->winobj || !parent)
+	if (!allowJS || !cf->winobj || !parent)
 		return 0;
 
 	debugPrint(5, "> new array %s", name);
@@ -983,7 +999,7 @@ static int set_array_element(jsobjtype array, int idx,
 {
 	int l;
 
-	if (!allowJS || !cw->winobj || !array)
+	if (!allowJS || !cf->winobj || !array)
 		return -1;
 
 	debugPrint(5, "> set [%d]=%s", idx, debugString(value));
@@ -1040,7 +1056,7 @@ jsobjtype instantiate(jsobjtype parent, const char *name, const char *classname)
 	if (whichproc == 'j')
 		return instantiate_nat(parent, name, classname);
 
-	if (!allowJS || !cw->winobj || !parent)
+	if (!allowJS || !cf->winobj || !parent)
 		return 0;
 
 	debugPrint(5, "> instantiate %s %s", name,
@@ -1088,7 +1104,7 @@ static int run_function(jsobjtype obj, const char *name, int argc,
 {
 	int rc;
 	propval = 0;		/* should already be 0 */
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return -1;
 
 	debugPrint(5, "> call %s(%d)", name, argc);
@@ -1148,6 +1164,18 @@ void update_var_in_js(int varid)
 		return;
 	if (varid == 1)
 		value = allowXHR;
+	if (varid == 2)
+		value = debugLevel;
+	if (varid == 3)
+		value = verifyCertificates;
+	if (varid == 4) {
+		int i;
+		for (i = 0; i < 10; ++i)
+			if (userAgents[i] == currentAgent) {
+				value = i;
+				break;
+			}
+	}
 	debugPrint(5, "> varupdate %d", varid);
 	head.cmd = EJ_CMD_VARUPDATE;
 	head.obj = 0;
@@ -1170,7 +1198,7 @@ void set_basehref(const char *h)
 {
 	if (!h)
 		h = emptyString;
-	set_property_string(cw->docobj, "base$href$", h);
+	set_property_string(cf->docobj, "base$href$", h);
 }				/* set_basehref */
 
 /* The object is a select-one field in the form, and this function returns
@@ -1182,7 +1210,7 @@ char *get_property_option(jsobjtype obj)
 	jsobjtype oo;		/* option object */
 	char *val;
 
-	if (!allowJS || !cw->winobj || !obj)
+	if (!allowJS || !cf->winobj || !obj)
 		return 0;
 
 	n = get_property_number(obj, "selectedIndex");
@@ -1209,7 +1237,7 @@ static void docCookie(jsobjtype d)
 {
 	int cook_l;
 	char *cook = initString(&cook_l);
-	const char *url = cw->fileName;
+	const char *url = cf->fileName;
 	bool secure = false;
 	const char *proto;
 	char *s;
@@ -1251,8 +1279,8 @@ int uname(struct utsname *pun)
  * and methods that are base for client side DOM. */
 void setupJavaDom(void)
 {
-	jsobjtype w = cw->winobj;	// window object
-	jsobjtype d = cw->docobj;	// document object
+	jsobjtype w = cf->winobj;	// window object
+	jsobjtype d = cf->docobj;	// document object
 	jsobjtype nav;		// navigator object
 	jsobjtype navpi;	// navigator plugins
 	jsobjtype navmt;	// navigator mime types
@@ -1326,7 +1354,7 @@ void setupJavaDom(void)
 	hist = instantiate(w, "history", 0);
 	if (hist == NULL)
 		return;
-	set_property_string(hist, "current", cw->fileName);
+	set_property_string(hist, "current", cf->fileName);
 /* Since there is no history in edbrowse, the rest is left to startwindow.js */
 
 /* the js window/document setup script.
@@ -1337,10 +1365,10 @@ void setupJavaDom(void)
 // Document properties that must be set after startwindow.js.
 // Most of these use the setters in the URL class.
 	set_property_string(d, "referrer", cw->referrer);
-	instantiate_url(d, "URL", cw->fileName);
-	instantiate_url(d, "location", cw->fileName);
-	instantiate_url(w, "location", cw->fileName);
-	set_property_string(d, "domain", getHostURL(cw->fileName));
+	instantiate_url(d, "URL", cf->fileName);
+	instantiate_url(d, "location", cf->fileName);
+	instantiate_url(w, "location", cf->fileName);
+	set_property_string(d, "domain", getHostURL(cf->fileName));
 	docCookie(d);
 }				/* setupJavaDom */
 
@@ -1391,11 +1419,9 @@ in a primary menu. These new options must map back to html tags,
 and then to the dropdown list as you interact with the form.
 This is tested in jsrt - select a state,
 whereupon the colors below, that you have to choose from, can change.
-
-Someday the entire page will be rerendered based upon the js tree,
-which could be modified in almost any way, but today I only look at
-changing menus, because that is the high runner case.
-Besides, it use to seg fault when I didn't watch for this.
+This does not easily fold into rerender(),
+since the line <> in the buffer looks exactly the same,
+so this tells you the options underneath have changed.
 *********************************************************************/
 
 static void rebuildSelector(struct htmlTag *sel, jsobjtype oa, int len2)
@@ -1524,6 +1550,9 @@ void rebuildSelectors(void)
 	jsobjtype oa;		/* option array */
 	int len;		/* length of option array */
 
+	if (!isJSAlive)
+		return;
+
 	for (i1 = 0; i1 < cw->numTags; ++i1) {
 		t = tagList[i1];
 		if (!t->jv)
