@@ -5,6 +5,7 @@ bool getUserPass(const char *url, char *creds, bool find_proxy) ;
 bool addWebAuthorization(const char *url, const char *credentials, bool proxy) ;
 
 /* sourcefile=buffers.c */
+void removeHiddenNumbers(pst p, uchar terminate);
 pst fetchLine(int n, int show) ;
 void displayLine(int n) ;
 void initializeReadline(void) ;
@@ -13,7 +14,7 @@ bool cxCompare(int cx) ;
 bool cxActive(int cx) ;
 bool cxQuit(int cx, int action) ;
 void cxSwitch(int cx, bool interactive) ;
-bool addTextToBuffer(const pst inbuf, int length, int destl, bool onside) ;
+bool addTextToBuffer(const pst inbuf, int length, int destl, bool showtrail) ;
 void delText(int start, int end) ;
 bool readFileArgv(const char *filename);
 bool unfoldBufferW(const struct ebWindow *w, bool cr, char **data, int *len) ;
@@ -33,6 +34,13 @@ bool receiveCookie(const char *url, const char *str) ;
 void cookiesFromJar(void) ;
 void sendCookies(char **s, int *l, const char *url, bool issecure) ;
 
+/* sourcefile=cache.c */
+void setupEdbrowseCache(void);
+void clearCache(void) ;
+bool fetchCache(const char * url, const char *etag, time_t modtime, char **data, int *data_len) ;
+void storeCache(const char *url, const char *etag, time_t modtime, const char *data, int datalen) ;
+bool presentInCache(const char *url) ;
+
 /* sourcefile=dbodbc.c (and others) */
 bool fetchForeign(char *tname) ;
 
@@ -54,7 +62,7 @@ int goSelect(int *startLine, char **rbuf) ;
 /* sourcefile=ebjs.c */
 void dwStart(void);
 void createJavaContext(void) ;
-void freeJavaContext(struct ebWindow *w) ;
+void freeJavaContext(struct ebFrame *f) ;
 void js_shutdown(void) ;
 void js_disconnect(void);
 char *jsRunScriptResult(jsobjtype obj, const char *str, const char *filename, int lineno) ;
@@ -107,8 +115,14 @@ void breakLineSetup(void) ;
 char *htmlReformat(char *buf) ;
 void extractEmailAddresses(char *line) ;
 void cutDuplicateEmails(char *tolist, char *cclist, const char *reply) ;
-bool looksBinary(const char *buf, int buflen) ;
-void looks_8859_utf8(const char *buf, int buflen, bool * iso_p, bool * utf8_p) ;
+int byteOrderMark(const uchar *buf, int buflen);
+bool looksBinary(const char *buf, int buflen);
+void looks_8859_utf8(const char *buf, int buflen, bool * iso_p, bool * utf8_p);
+void iso2utf(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p);
+void utf2iso(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p);
+void utfHigh(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p, bool inutf8, bool out32, bool outbig);
+char *uni2utf8(unsigned int unichar);
+void utfLow(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p, int bom);
 char *base64Encode(const char *inbuf, int inlen, bool lines);
 void iuReformat(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p) ;
 bool parseDataURI(const char *uri, char **mediatype, char **data, int *data_l);
@@ -116,8 +130,10 @@ bool parseDataURI(const char *uri, char **mediatype, char **data, int *data_l);
 /* sourcefile=html.c */
 bool tagHandler(int seqno, const char *name) ;
 void jSideEffects(void) ;
-void jSyncup(void) ;
+void jSyncup(bool fromtimer) ;
+void jClearSync(void);
 void htmlMetaHelper(struct htmlTag *t);
+extern void runScriptsPending(void);
 void preFormatCheck(int tagno, bool * pretag, bool * slash) ;
 char *htmlParse(char *buf, int remote) ;
 bool htmlTest(void) ;
@@ -131,8 +147,10 @@ bool handlerGoBrowse(const struct htmlTag *t, const char *name) ;
 void runningError(int msg, ...) ;
 void rerender(bool notify);
 void delTags(int startRange, int endRange);
+extern void runOnload(void);
 bool timerWait(int *delay_sec, int *delay_ms);
-void delTimers(struct ebWindow *w);
+void delTimers(struct ebFrame *f);
+void delInputChanges(struct ebFrame *f);
 void runTimers(void);
 void javaOpensWindow(const char *href, const char *name) ;
 void javaSetsLinkage(bool after, char type, jsobjtype p, const char *rest);
@@ -166,7 +184,7 @@ int base64Decode(char *start, char **end);
 char *extractHeaderParam(const char *str, const char *item) ;
 time_t parseHeaderDate(const char *date) ;
 bool parseRefresh(char *ref, int *delay_p) ;
-bool refreshDelay(int sec, const char *u) ;
+bool shortRefreshDelay(void);
 bool httpConnect(const char *url, bool down_ok, bool webpage, bool f_encoded, char **headers_p, char **body_p, int *bodlen_p);
 void ebcurl_setError(CURLcode curlret, const char *url) ;
 void setHTTPLanguage(const char *lang) ;
@@ -176,6 +194,9 @@ void addNovsHost(char *host) ;
 void deleteNovsHosts(void);
 CURLcode setCurlURL(CURL * h, const char *url) ;
 const char *findProxyForURL(const char *url) ;
+bool frameExpand(bool expand, int ln1, int ln2);
+struct htmlTag *line2frame(int ln);
+bool reexpandFrame(void);
 
 /* sourcefile=main.c */
 const char *mailRedirect(const char *to, const char *from, const char *reply, const char *subj) ;
@@ -210,8 +231,6 @@ bool validAccount(int n) ;
 bool sendMailCurrent(int sm_account, bool dosig) ;
 
 /* sourcefile=messages.c */
-void iso2utf(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p) ;
-void utf2iso(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p) ;
 void selectLanguage(void) ;
 const char *i_getString(int msg);
 void i_puts(int msg) ;
