@@ -101,7 +101,7 @@ static void writeAttachment(struct MHINFO *w)
 			    ((pst) w->start, w->end - w->start, 0, false))
 				i_printf(MSG_AttNoCopy, cx);
 			else if (w->cfn[0])
-				cw->fileName = cloneString(w->cfn);
+				cf->fileName = cloneString(w->cfn);
 			cxSwitch(svcx, false);	/* back to where we were */
 		}
 	} else if (!stringEqual(atname, "x")) {
@@ -1303,9 +1303,10 @@ static char presentMail(void)
 			key = 'w';
 			if (*redirect == '-')
 				++redirect, key = 'u';
-			if (stringEqual(redirect, "x"))
-				i_puts(MSG_Junk);
-			else
+			if (stringEqual(redirect, "x")) {
+				i_printf(MSG_Junk);
+				printf("[%s]\n", lastMailInfo->subject);
+			} else
 				printf("> %s\n", redirect);
 		} else {
 			if (*redirect == '-')
