@@ -107,6 +107,7 @@ static void anchorSwap(char *buf)
 	char c, d, *s, *ss, *w, *a;
 	bool pretag;		// <pre>
 	bool premode;		// inside <pre> </pre>
+	bool inputmode;		// inside an input field
 	bool slash;		// closing tag
 	bool change;		// made a swap somewhere
 	bool strong;		// strong whitespace, newline or paragraph
@@ -123,10 +124,28 @@ static void anchorSwap(char *buf)
  * Watch out for utf8 - don't translate the a0 in c3a0.  That is a grave.
  * But a0 by itself is breakspace; turn it into space.
  * And c2a0 is a0 is breakspace.
- * Then get rid of hyperlinks with absolutely nothing to click on. */
+ * Don't do any of these transliterations in an input field. */
 
+	inputmode = false;
 	for (s = w = buf; c = *s; ++s) {
 		d = s[1];
+		if (c == InternalCodeChar && isdigitByte(d)) {
+			strtol(s + 1, &ss, 10);
+			if (*ss == '<')
+				inputmode = true;
+			if (*ss == '>')
+				inputmode = false;
+			++ss;
+			n = ss - s;
+			memmove(w, s, n);
+			w += n;
+			s = ss - 1;
+			continue;
+		}
+
+		if (inputmode)
+			goto put1;
+
 /* utf8 test */
 		if ((c & 0xc0) == 0xc0 && (d & 0xc0) == 0x80) {
 			unsigned int uni = 0;
@@ -152,7 +171,7 @@ static void anchorSwap(char *buf)
 			continue;
 		}
 
-/* Now assuming iso8859-1, which is practically depricated */
+/* Now assuming iso8859-1, which is practically deprecated */
 		ss = strchr(from, c);
 		if (ss)
 			c = becomes[ss - from];
@@ -741,6 +760,7 @@ char *htmlReformat(char *buf)
 	bool pretag, slash;
 	char *new;
 	int l, tagno, extra;
+	char *fmark;		/* mark the start of a frame */
 
 	cellDelimiters(buf);
 
@@ -830,48 +850,15 @@ char *htmlReformat(char *buf)
 		l += 4;
 	}
 
-	return new;
-}				/* htmlReformat */
-
-/*********************************************************************
-Convert a 31 bit unicode character into utf8.
-*********************************************************************/
-
-static void uni2utf8(unsigned int unichar, unsigned char *outbuf)
-{
-	int n = 0;
-
-	if (unichar <= 0x7f) {
-		outbuf[n++] = unichar;
-	} else if (unichar <= 0x7ff) {
-		outbuf[n++] = 0xc0 | ((unichar >> 6) & 0x1f);
-		outbuf[n++] = 0x80 | (unichar & 0x3f);
-	} else if (unichar <= 0xffff) {
-		outbuf[n++] = 0xe0 | ((unichar >> 12) & 0xf);
-		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
-		outbuf[n++] = 0x80 | (unichar & 0x3f);
-	} else if (unichar <= 0x1fffff) {
-		outbuf[n++] = 0xf0 | ((unichar >> 18) & 7);
-		outbuf[n++] = 0x80 | ((unichar >> 12) & 0x3f);
-		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
-		outbuf[n++] = 0x80 | (unichar & 0x3f);
-	} else if (unichar <= 0x3ffffff) {
-		outbuf[n++] = 0xf8 | ((unichar >> 24) & 3);
-		outbuf[n++] = 0x80 | ((unichar >> 18) & 0x3f);
-		outbuf[n++] = 0x80 | ((unichar >> 12) & 0x3f);
-		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
-		outbuf[n++] = 0x80 | (unichar & 0x3f);
-	} else if (unichar <= 0x7fffffff) {
-		outbuf[n++] = 0xfc | ((unichar >> 30) & 1);
-		outbuf[n++] = 0x80 | ((unichar >> 24) & 0x3f);
-		outbuf[n++] = 0x80 | ((unichar >> 18) & 0x3f);
-		outbuf[n++] = 0x80 | ((unichar >> 12) & 0x3f);
-		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
-		outbuf[n++] = 0x80 | (unichar & 0x3f);
+/* It's a little thing really, but the blank line at the top of each frame annoys me */
+	fmark = new;
+	while (fmark = strstr(fmark + 1, "*`--\n\n")) {
+		if (isdigit(fmark[-1]))
+			strmove(fmark + 5, fmark + 6);
 	}
 
-	outbuf[n] = 0;
-}				/* uni2utf8 */
+	return new;
+}				/* htmlReformat */
 
 /*********************************************************************
 Crunch a to-list or a copy-to-list down to its email addresses.
@@ -1004,13 +991,32 @@ void cutDuplicateEmails(char *tolist, char *cclist, const char *reply)
 	}
 }				/* cutDuplicateEmails */
 
+/* return 1 for utf16, 2 for utf32, ored with 4 for big endian */
+int byteOrderMark(const uchar * buf, int buflen)
+{
+	if (buflen < 2)
+		return 0;
+	if (buf[0] == 0xfe && buf[1] == 0xff)
+		return 5;
+	if (buf[0] == 0xff && buf[1] == 0xfe) {
+		if (buflen >= 4 && buf[2] == 0 && buf[3] == 0)
+			return 2;
+		return 1;
+	}
+	if (buflen >= 4 && !memcmp(buf, "\x0\x0\xfe\xff", 4))
+		return 6;
+	return 0;
+}				/* byteOrderMark */
+
 /*********************************************************************
-We got some data, from a file or from the internet.
+We got some data from a file or from the internet.
 Count the binary characters and decide if this is, on the whole,
 binary or text.  I allow some nonascii chars,
 like you might see in Spanish or German, and still call it text,
 but if there's too many such chars, I call it binary.
 It's not an exact science.
+utf8 sequences are considered text characters.
+If there is a leading byte order mark as per the previous routine, it's text.
 *********************************************************************/
 
 bool looksBinary(const char *buf, int buflen)
@@ -1018,6 +1024,10 @@ bool looksBinary(const char *buf, int buflen)
 	int i, j, bincount = 0, charcount = 0;
 	char c;
 	uchar seed;
+
+	if (byteOrderMark((uchar *) buf, buflen))
+		return false;
+
 	for (i = 0; i < buflen; ++i, ++charcount) {
 		c = buf[i];
 // 0 is ascii, but not really text, and very common in binary files.
@@ -1042,6 +1052,7 @@ binchar:
 // this is valid utf8 char, don't treat it as binary.
 		i += j;
 	}
+
 	return (bincount * 8 - 16 >= charcount);
 }				/* looksBinary */
 
@@ -1084,6 +1095,395 @@ isogo:
 		*iso_p = true;
 }				/* looks_8859_utf8 */
 
+/*********************************************************************
+Convert a string from iso 8859 to utf8, or vice versa.
+In each case a new string is allocated.
+Don't forget to free it when you're done.
+*********************************************************************/
+
+/* only 8859-1 and 8859-2 so far */
+static const int iso_unicodes[2][128] = {
+	{0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b,
+	 0x8c, 0x8d, 0x8e, 0x8f,
+	 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a,
+	 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
+	 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
+	 0xab, 0xac, 0xad, 0xae, 0xaf,
+	 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba,
+	 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
+	 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca,
+	 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
+	 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda,
+	 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
+	 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
+	 0xeb, 0xec, 0xed, 0xee, 0xef,
+	 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa,
+	 0xfb, 0xfc, 0xfd, 0xfe, 0xff},
+	{0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b,
+	 0x8c, 0x8d, 0x8e, 0x8f,
+	 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a,
+	 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
+	 0xa0, 0x104, 0x2d8, 0x141, 0xa4, 0x13d, 0x15a, 0xa7, 0xa8, 0x160,
+	 0x15e, 0x164, 0x179, 0xad, 0x17d, 0x17b,
+	 0xb0, 0x105, 0x2db, 0x142, 0xb4, 0x13e, 0x15b, 0x2c7, 0xb8, 0x161,
+	 0x15f, 0x165, 0x17a, 0x2dd, 0x17e, 0x17c,
+	 0x154, 0xc1, 0xc2, 0x102, 0xc4, 0x139, 0x106, 0xc7, 0x10c, 0xc9,
+	 0x118, 0xcb, 0x11a, 0xcd, 0xce, 0x10e,
+	 0x110, 0x143, 0x147, 0xd3, 0xd4, 0x150, 0xd6, 0xd7, 0x158, 0x16e,
+	 0xda, 0x170, 0xdc, 0xdd, 0x162, 0xdf,
+	 0x155, 0xe1, 0xe2, 0x103, 0xe4, 0x13a, 0x107, 0xe7, 0x10d, 0xe9,
+	 0x119, 0xeb, 0x11b, 0xed, 0xee, 0x10f,
+	 0x111, 0x144, 0x148, 0xf3, 0xf4, 0x151, 0xf6, 0xf7, 0x159, 0x16f,
+	 0xfa, 0x171, 0xfc, 0xfd, 0x163, 0x2d9},
+};
+
+void iso2utf(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p)
+{
+	int i, j;
+	int nacount = 0;
+	char c;
+	char *outbuf;
+	const int *isoarray = iso_unicodes[type8859 - 1];
+	int ucode;
+
+	if (!inbuflen) {
+		*outbuf_p = emptyString;
+		*outbuflen_p = 0;
+		return;
+	}
+
+/* count chars, so we can allocate */
+	for (i = 0; i < inbuflen; ++i) {
+		c = inbuf[i];
+		if (c < 0)
+			++nacount;
+	}
+
+	outbuf = allocString(inbuflen + nacount + 1);
+	for (i = j = 0; i < inbuflen; ++i) {
+		c = inbuf[i];
+		if (c >= 0) {
+			outbuf[j++] = c;
+			continue;
+		}
+		ucode = isoarray[c & 0x7f];
+		outbuf[j++] = (ucode >> 6) | 0xc0;
+		outbuf[j++] = (ucode & 0x3f) | 0x80;
+	}
+	outbuf[j] = 0;		/* just for fun */
+
+	*outbuf_p = outbuf;
+	*outbuflen_p = j;
+}				/* iso2utf */
+
+void utf2iso(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p)
+{
+	int i, j, k;
+	char c;
+	char *outbuf;
+	const int *isoarray = iso_unicodes[type8859 - 1];
+	int ucode;
+
+	if (!inbuflen) {
+		*outbuf_p = emptyString;
+		*outbuflen_p = 0;
+		return;
+	}
+
+	outbuf = allocString(inbuflen + 1);
+	for (i = j = 0; i < inbuflen; ++i) {
+		c = inbuf[i];
+
+/* regular chars and nonascii chars that aren't utf8 pass through. */
+/* There shouldn't be any of the latter */
+		if (((uchar) c & 0xc0) != 0xc0) {
+			outbuf[j++] = c;
+			continue;
+		}
+
+/* Convertable into 11 bit */
+		if (((uchar) c & 0xe0) == 0xc0
+		    && ((uchar) inbuf[i + 1] & 0xc0) == 0x80) {
+			ucode = c & 0x1f;
+			ucode <<= 6;
+			ucode |= (inbuf[i + 1] & 0x3f);
+			for (k = 0; k < 128; ++k)
+				if (isoarray[k] == ucode)
+					break;
+			if (k < 128) {
+				outbuf[j++] = k | 0x80;
+				++i;
+				continue;
+			}
+		}
+
+/* unicodes not found in our iso class are converted into stars */
+		c <<= 1;
+		++i;
+		for (++i; c < 0; ++i, c <<= 1) {
+			if (((uchar) outbuf[i] & 0xc0) != 0x80)
+				break;
+		}
+		outbuf[j++] = '*';
+		--i;
+	}
+	outbuf[j] = 0;		/* just for fun */
+
+	*outbuf_p = outbuf;
+	*outbuflen_p = j;
+}				/* utf2iso */
+
+/*********************************************************************
+Convert the current line in buffer, which is either iso8859-1 or utf8,
+into utf16 or utf32, big or little endian.
+The returned string is allocated, though not really a string,
+since it will contain nulls, plenty of them in the case of utf32.
+*********************************************************************/
+
+void utfHigh(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p,
+	     bool inutf8, bool out32, bool outbig)
+{
+	uchar *outbuf;
+	unsigned int unicode;
+	uchar c;
+	int i, j;
+
+	if (!inbuflen) {
+		*outbuf_p = emptyString;
+		*outbuflen_p = 0;
+		return;
+	}
+
+	outbuf = allocMem(inbuflen * 4);	// worst case
+
+	i = j = 0;
+	while (i < inbuflen) {
+		c = (uchar) inbuf[i];
+		if (!inutf8 || (c & 0xc0) != 0xc0 && (c & 0xfe) != 0xfe) {
+			unicode = c;	// that was easy
+			++i;
+		} else {
+			uchar mask = 0x20;
+			int k = 1;
+			++i;
+			while (c & mask)
+				++k, mask >>= 1;
+			c &= (mask - 1);
+			unicode = ((unsigned int)c) << (6 * k);
+			while (i < inbuflen && k) {
+				c = (uchar) inbuf[i];
+				if ((c & 0xc0) != 0x80)
+					break;
+				++i, --k;
+				c &= 0x3f;
+				unicode |= (((unsigned int)c) << (6 * k));
+			}
+		}
+
+		if (out32) {
+			if (outbig) {
+				outbuf[j++] = ((unicode >> 24) & 0xff);
+				outbuf[j++] = ((unicode >> 16) & 0xff);
+				outbuf[j++] = ((unicode >> 8) & 0xff);
+				outbuf[j++] = (unicode & 0xff);
+			} else {
+				outbuf[j++] = (unicode & 0xff);
+				outbuf[j++] = ((unicode >> 8) & 0xff);
+				outbuf[j++] = ((unicode >> 16) & 0xff);
+				outbuf[j++] = ((unicode >> 24) & 0xff);
+			}
+			continue;
+		}
+// utf16, a bit trickier but not too bad.
+		if (unicode <= 0xd7ff || unicode >= 0xe000 && unicode <= 0xffff) {
+			if (outbig) {
+				outbuf[j++] = ((unicode >> 8) & 0xff);
+				outbuf[j++] = (unicode & 0xff);
+			} else {
+				outbuf[j++] = (unicode & 0xff);
+				outbuf[j++] = ((unicode >> 8) & 0xff);
+			}
+			continue;
+		}
+
+		if (unicode >= 0x10000 && unicode <= 0x10ffff) {
+// surrogate pairs
+			unsigned int pair1, pair2;
+			unicode -= 0x10000;
+			pair1 = 0xd800 + ((unicode >> 10) & 0x3ff);
+			pair2 = 0xdc00 + (unicode & 0x3ff);
+			if (outbig) {
+				outbuf[j++] = ((pair1 >> 8) & 0xff);
+				outbuf[j++] = (pair1 & 0xff);
+				outbuf[j++] = ((pair2 >> 8) & 0xff);
+				outbuf[j++] = (pair2 & 0xff);
+			} else {
+				outbuf[j++] = (pair1 & 0xff);
+				outbuf[j++] = ((pair1 >> 8) & 0xff);
+				outbuf[j++] = (pair2 & 0xff);
+				outbuf[j++] = ((pair2 >> 8) & 0xff);
+			}
+			continue;
+		}
+
+	}
+
+	*outbuf_p = (char *)outbuf;
+	*outbuflen_p = j;
+}				/* utfHigh */
+
+/* convert a 32 bit unicode character into utf8 */
+char *uni2utf8(unsigned int unichar)
+{
+	static uchar outbuf[12];
+	int n = 0;
+
+	if (unichar <= 0x7f) {
+		outbuf[n++] = unichar;
+	} else if (unichar <= 0x7ff) {
+		outbuf[n++] = 0xc0 | ((unichar >> 6) & 0x1f);
+		outbuf[n++] = 0x80 | (unichar & 0x3f);
+	} else if (unichar <= 0xffff) {
+		outbuf[n++] = 0xe0 | ((unichar >> 12) & 0xf);
+		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
+		outbuf[n++] = 0x80 | (unichar & 0x3f);
+	} else if (unichar <= 0x1fffff) {
+		outbuf[n++] = 0xf0 | ((unichar >> 18) & 7);
+		outbuf[n++] = 0x80 | ((unichar >> 12) & 0x3f);
+		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
+		outbuf[n++] = 0x80 | (unichar & 0x3f);
+	} else if (unichar <= 0x3ffffff) {
+		outbuf[n++] = 0xf8 | ((unichar >> 24) & 3);
+		outbuf[n++] = 0x80 | ((unichar >> 18) & 0x3f);
+		outbuf[n++] = 0x80 | ((unichar >> 12) & 0x3f);
+		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
+		outbuf[n++] = 0x80 | (unichar & 0x3f);
+	} else if (unichar <= 0x7fffffff) {
+		outbuf[n++] = 0xfc | ((unichar >> 30) & 1);
+		outbuf[n++] = 0x80 | ((unichar >> 24) & 0x3f);
+		outbuf[n++] = 0x80 | ((unichar >> 18) & 0x3f);
+		outbuf[n++] = 0x80 | ((unichar >> 12) & 0x3f);
+		outbuf[n++] = 0x80 | ((unichar >> 6) & 0x3f);
+		outbuf[n++] = 0x80 | (unichar & 0x3f);
+	}
+
+	outbuf[n] = 0;
+	return (char *)outbuf;
+}				/* uni2utf8 */
+
+void utfLow(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p,
+	    int bom)
+{
+	char *obuf;
+	int obuf_l;
+	unsigned int unicode;
+	int isbig;
+	int k, l;
+	const int *isoarray = iso_unicodes[type8859 - 1];
+
+	if (!inbuflen) {
+		*outbuf_p = emptyString;
+		*outbuflen_p = 0;
+		return;
+	}
+
+	obuf = initString(&obuf_l);
+	isbig = (bom & 4);
+	bom &= 3;
+	l = bom * 2;		// skip past byte order mark
+
+	while (l < inbuflen) {
+		if (bom == 2) {
+			if (l + 4 > inbuflen) {
+				unicode = '?';
+				l = inbuflen;
+			} else if (isbig) {
+				unicode = (uchar) inbuf[l];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l + 1];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l + 2];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l + 3];
+				l += 4;
+			} else {
+				unicode = (uchar) inbuf[l + 3];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l + 2];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l + 1];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l];
+				l += 4;
+			}
+		} else {
+			if (l + 2 > inbuflen) {
+				unicode = '?';
+				l = inbuflen;
+			} else if (isbig) {
+				unicode = (uchar) inbuf[l];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l + 1];
+				l += 2;
+			} else {
+				unicode = (uchar) inbuf[l + 1];
+				unicode <<= 8;
+				unicode |= (uchar) inbuf[l];
+				l += 2;
+			}
+			if (unicode >= 0xd800 && unicode <= 0xdbff
+			    && l + 2 <= inbuflen) {
+				unsigned int pair1, pair2;
+				pair1 = unicode - 0xd800;
+				if (isbig) {
+					pair2 = (uchar) inbuf[l];
+					pair2 <<= 8;
+					pair2 |= (uchar) inbuf[l + 1];
+				} else {
+					pair2 = (uchar) inbuf[l + 1];
+					pair2 <<= 8;
+					pair2 |= (uchar) inbuf[l];
+				}
+				if (pair2 >= 0xdc00 && pair2 <= 0xdfff) {
+					pair2 -= 0xdc00;
+					l += 2;
+					unicode = pair1;
+					unicode <<= 10;
+					unicode |= pair2;
+				}
+			}
+		}
+
+// ok we got the unicode.
+// It now becomes utf8 or iso8859-x
+		if (cons_utf8) {
+			stringAndString(&obuf, &obuf_l, uni2utf8(unicode));
+			continue;
+		}
+// iso8859-x here, practically deprecated
+		if (unicode <= 127) {	// ascii
+			stringAndChar(&obuf, &obuf_l, (char)unicode);
+			continue;
+		}
+
+		for (k = 0; k < 128; ++k)
+			if (isoarray[k] == unicode)
+				break;
+		if (k < 128)
+			unicode = k | 0x80;
+		else
+			unicode = '?';
+		stringAndChar(&obuf, &obuf_l, (char)unicode);
+	}
+
+// The input string is a file or url and has 2 extra bytes after it.
+// After reformatting it should still have two extra bytes after it.
+	stringAndString(&obuf, &obuf_l, "  ");
+
+	*outbuf_p = obuf;
+	*outbuflen_p = obuf_l - 2;
+}				/* utfLow */
+
 static char base64_chars[] =
     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
