@@ -38,22 +38,38 @@ struct listHead down_jobs = {
 
 static void background_download(struct eb_curl_callback_data *data);
 static void setup_download(struct eb_curl_callback_data *data);
-static char errorText[CURL_ERROR_SIZE + 1];
-static char *http_headers;
-static int http_headers_len;
-static char *httpLanguage;
-/* http content type is used in many places, and isn't arbitrarily long
- * or case sensitive, so keep our own sanitized copy. */
-static char hct[60];
-static char *hct2;		/* extra content info such as charset */
-int hcl;			/* http content length */
 static CURL *http_curl_init(struct eb_curl_callback_data *cbd);
 
-/* This function is called for web redirection, by the refresh command,
- * or by window.location = new_url. */
+static char *http_headers;
+static int http_headers_len;
+static char *httpLanguage;	/* outgoing */
+long ht_code;			/* example, 404 */
+static char ht_error[CURL_ERROR_SIZE + 1];
+/* an assortment of variables that are gleaned from the incoming http headers */
+/* http content type is used in many places, and isn't arbitrarily long
+ * or case sensitive, so keep our own sanitized copy. */
+static char ht_content[60];
+static char *ht_charset;	/* extra content info such as charset */
+int ht_length;			/* http content length */
+static char *ht_cdfn;		/* http content disposition file name */
+static time_t ht_modtime;	/* http modification time */
+static char *ht_etag;		/* the etag in the header */
+static bool ht_cacheable;
+
+/*********************************************************************
+This function is called for a new web page, by http refresh,
+or by http redirection,
+or by document.location = new_url, or by new Window().
+If delay is 0 or less then the action should happen now.
+-1 is from http header location:
+The refresh parameter means replace the current page.
+This is false only if js creates a new window, which should stack up on top of the old.
+*********************************************************************/
+
 char *newlocation;
 int newloc_d;			/* possible delay */
 bool newloc_r;			/* replace the buffer */
+struct ebFrame *newloc_f;	/* frame calling for new web page */
 bool js_redirects;
 void gotoLocation(char *url, int delay, bool rf)
 {
@@ -65,6 +81,7 @@ void gotoLocation(char *url, int delay, bool rf)
 	newlocation = url;
 	newloc_d = delay;
 	newloc_r = rf;
+	newloc_f = cf;
 	if (!delay)
 		js_redirects = true;
 }				/* gotoLocation */
@@ -120,24 +137,71 @@ static void scan_http_headers(bool fromCallback)
 {
 	char *v;
 
-	if (!hct[0] && (v = find_http_header("content-type"))) {
-		strncpy(hct, v, sizeof(hct) - 1);
-		caseShift(hct, 'l');
+	if (!ht_content[0] && (v = find_http_header("content-type"))) {
+		strncpy(ht_content, v, sizeof(ht_content) - 1);
+		caseShift(ht_content, 'l');
 		nzFree(v);
-		debugPrint(3, "content %s", hct);
-		hct2 = strchr(hct, ';');
-		if (hct2)
-			*hct2++ = 0;
+		debugPrint(3, "content %s", ht_content);
+		ht_charset = strchr(ht_content, ';');
+		if (ht_charset)
+			*ht_charset++ = 0;
 /* The protocol, such as rtsp, could have already set the mime type. */
-		if (!cw->mt)
-			cw->mt = findMimeByContent(hct);
+		if (!cf->mt)
+			cf->mt = findMimeByContent(ht_content);
 	}
 
-	if (!hcl && (v = find_http_header("content-length"))) {
-		hcl = atoi(v);
+	if (!ht_cdfn && (v = find_http_header("content-disposition"))) {
+		char *s = strstrCI(v, "filename=");
+		if (s) {
+			s += 9;
+			if (*s == '"') {
+				char *t;
+				++s;
+				t = strchr(s, '"');
+				if (t)
+					*t = 0;
+			}
+			ht_cdfn = cloneString(s);
+			debugPrint(3, "disposition filename %s", ht_cdfn);
+		}
+		nzFree(v);
+	}
+
+	if (!ht_length && (v = find_http_header("content-length"))) {
+		ht_length = atoi(v);
+		nzFree(v);
+		if (ht_length)
+			debugPrint(3, "content length %d", ht_length);
+	}
+
+	if (!ht_etag && (v = find_http_header("etag"))) {
+		ht_etag = v;
+		debugPrint(3, "etag %s", ht_etag);
+	}
+
+	if (ht_cacheable && (v = find_http_header("cache-control"))) {
+		caseShift(v, 'l');
+		if (strstr(v, "no-cache")) {
+			ht_cacheable = false;
+			debugPrint(3, "no cache");
+		}
+		nzFree(v);
+	}
+
+	if (ht_cacheable && (v = find_http_header("pragma"))) {
+		caseShift(v, 'l');
+		if (strstr(v, "no-cache")) {
+			ht_cacheable = false;
+			debugPrint(3, "no cache");
+		}
+		nzFree(v);
+	}
+
+	if (!ht_modtime && (v = find_http_header("last-modified"))) {
+		ht_modtime = parseHeaderDate(v);
+		if (ht_modtime)
+			debugPrint(3, "mod date %s", v);
 		nzFree(v);
-		if (hcl)
-			debugPrint(3, "content length %d", hcl);
 	}
 
 	if (fromCallback)
@@ -145,8 +209,7 @@ static void scan_http_headers(bool fromCallback)
 
 	if ((newlocation == NULL) && (v = find_http_header("location"))) {
 		unpercentURL(v);
-		newlocation = v;
-		newloc_d = -1;
+		gotoLocation(v, -1, true);
 	}
 
 	if (v = find_http_header("refresh")) {
@@ -162,7 +225,8 @@ static void scan_http_headers(bool fromCallback)
 }				/* scan_http_headers */
 
 /* actually run the curl request, http or ftp or whatever */
-static CURLcode fetch_internet(CURL * h, bool is_http)
+static bool is_http;
+static CURLcode fetch_internet(CURL * h)
 {
 	CURLcode curlret;
 	down_h = h;
@@ -171,9 +235,14 @@ static CURLcode fetch_internet(CURL * h, bool is_http)
 	newlocation = NULL;
 	nzFree(http_headers);
 	http_headers = initString(&http_headers_len);
-	hct[0] = 0;
-	hct2 = NULL;
-	hcl = 0;
+	ht_content[0] = 0;
+	ht_charset = NULL;
+	nzFree(ht_cdfn);
+	ht_cdfn = NULL;
+	nzFree(ht_etag);
+	ht_etag = NULL;
+	ht_length = 0;
+	ht_modtime = 0;
 	curlret = curl_easy_perform(h);
 	if (is_http)
 		scan_http_headers(false);
@@ -204,17 +273,24 @@ eb_curl_callback(char *incoming, size_t size, size_t nitems,
 	size_t num_bytes = nitems * size;
 	int dots1, dots2, rc;
 
+	if (data->down_state == 1 && is_http) {
+/* don't do a download unless the code is 200. */
+		curl_easy_getinfo(down_h, CURLINFO_RESPONSE_CODE, &ht_code);
+		if (ht_code != 200)
+			data->down_state = 0;
+	}
+
 	if (data->down_state == 1) {
-		if (hcl == 0) {
+		if (ht_length == 0) {
 // http should always set http content length, this is just for ftp.
 // And ftp downloading a file always has state = 1 on the first data block.
 			double d_size = 0.0;	// download size, if we can get it
 			curl_easy_getinfo(down_h,
 					  CURLINFO_CONTENT_LENGTH_DOWNLOAD,
 					  &d_size);
-			hcl = d_size;
-			if (hcl < 0)
-				hcl = 0;
+			ht_length = d_size;
+			if (ht_length < 0)
+				ht_length = 0;
 		}
 
 /* state 1, first data block, ask the user */
@@ -263,9 +339,9 @@ showdots:
 				putchar('.');
 			fflush(stdout);
 		}
-		if (showProgress == 'c' && hcl)
+		if (showProgress == 'c' && ht_length)
 			printf("%d/%d\n", dots2,
-			       (hcl + CHUNKSIZE - 1) / CHUNKSIZE);
+			       (ht_length + CHUNKSIZE - 1) / CHUNKSIZE);
 	}
 	return num_bytes;
 }
@@ -485,9 +561,21 @@ time_t parseHeaderDate(const char *date)
 	};
 	time_t t = 0;
 	int zone = 0;
+	time_t now = 0;
+	time_t utcnow = 0;
 	int y;			/* remember the type of format */
+	struct tm *temptm = NULL;
 	struct tm tm;
 	memset(&tm, 0, sizeof(struct tm));
+	tm.tm_isdst = -1;
+
+	now = time(NULL);
+	temptm = gmtime(&now);
+	if (temptm == NULL)
+		goto fail;
+	utcnow = mktime(temptm);
+	if (utcnow == -1 && errno)
+		goto fail;
 
 /* skip past day of the week */
 	date = strchr(date, ' ');
@@ -640,7 +728,7 @@ f2:
 
 	t = mktime(&tm);
 	if (t != (time_t) - 1)
-		return t + zone;
+		return t + zone + (now - utcnow);
 
 fail:
 	return 0;
@@ -672,7 +760,7 @@ bool parseRefresh(char *ref, int *delay_p)
 			u[-1] = 0;
 		debugPrint(3, "delay %d %s", delay, ref);
 /* avoid the obvious infinite loop */
-		if (sameURL(ref, cw->fileName)) {
+		if (sameURL(ref, cf->fileName)) {
 			*delay_p = 0;
 			return false;
 		}
@@ -686,17 +774,15 @@ bool parseRefresh(char *ref, int *delay_p)
 
 /* Return true if we waited for the duration, false if interrupted.
  * I don't know how to do this in Windows. */
-bool refreshDelay(int sec, const char *u)
+bool shortRefreshDelay(void)
 {
-/* the value 15 seconds is somewhat arbitrary */
-	if (sec < 15)
+/* the value 10 seconds is somewhat arbitrary */
+	if (newloc_d < 10)
 		return true;
-	i_printf(MSG_RedirectDelayed, u, sec);
+	i_printf(MSG_RedirectDelayed, newlocation, newloc_d);
 	return false;
-}				/* refreshDelay */
+}				/* shortRefreshDelay */
 
-long hcode;			/* example, 404 */
-static char herror[32];		/* example, file not found */
 static char *urlcopy;
 static int urlcopy_l;
 
@@ -739,12 +825,15 @@ static void urlSanitize(const char *url, const char *post, bool f_encoded)
 bool httpConnect(const char *url, bool down_ok, bool webpage,
 		 bool f_encoded, char **headers_p, char **body_p, int *bodlen_p)
 {
+	char *cacheData = NULL;
+	int cacheDataLen = 0;
 	CURL *h;		// the curl http handle
 	struct eb_curl_callback_data cbd;
 	char *referrer = NULL;
 	CURLcode curlret = CURLE_OK;
 	struct curl_slist *custom_headers = NULL;
 	struct curl_slist *tmp_headers = NULL;
+	const struct MIMETYPE *mt;
 	char user[MAXUSERPASS], pass[MAXUSERPASS];
 	char creds_buf[MAXUSERPASS * 2 + 1];	/* creds abr. for credentials */
 	int creds_len = 0;
@@ -759,6 +848,8 @@ bool httpConnect(const char *url, bool down_ok, bool webpage,
 	bool proceed_unauthenticated = false;
 	int redirect_count = 0;
 	bool name_changed = false;
+	bool post_request = false;
+	bool head_request = false;
 
 	if (headers_p)
 		*headers_p = 0;
@@ -771,7 +862,13 @@ bool httpConnect(const char *url, bool down_ok, bool webpage,
 	serverData = NULL;
 	serverDataLen = 0;
 	strcpy(creds_buf, ":");	/* Flush stale username and password. */
-	cw->mt = NULL;		/* should already be null */
+	cf->mt = NULL;		/* should already be null */
+
+	host = getHostURL(url);
+	if (!host) {
+		setError(MSG_DomainEmpty);
+		return false;
+	}
 
 /* Pull user password out of the url */
 	user[0] = pass[0] = 0;
@@ -805,10 +902,18 @@ bool httpConnect(const char *url, bool down_ok, bool webpage,
 		   stringEqualCI(prot, "scp") ||
 		   stringEqualCI(prot, "tftp") || stringEqualCI(prot, "sftp")) {
 		return ftpConnect(url, user, pass, f_encoded);
-	} else if ((cw->mt = findMimeByProtocol(prot)) && pluginsOn
-		   && cw->mt->stream) {
+	} else if ((cf->mt = findMimeByProtocol(prot)) && pluginsOn
+		   && cf->mt->stream) {
 mimestream:
-		cmd = pluginCommand(cw->mt, url, NULL, NULL);
+/* set this to null so we don't push a new buffer */
+		nzFree(serverData);
+		serverData = NULL;
+		serverDataLen = 0;
+/* could jump back here after a redirect, so use urlcopy if it is there */
+		cmd = pluginCommand(cf->mt, (urlcopy ? urlcopy : url),
+				    NULL, NULL);
+		nzFree(urlcopy);
+		urlcopy = NULL;
 		if (!cmd)
 			return false;
 		eb_system(cmd, true);
@@ -820,9 +925,16 @@ mimestream:
 	}
 
 /* Ok, it's http, but the suffix could force a plugin */
-	if ((cw->mt = findMimeByURL(url)) && pluginsOn && cw->mt->stream)
+	if ((cf->mt = findMimeByURL(url)) && pluginsOn && cf->mt->stream)
 		goto mimestream;
 
+/* if invoked from a playlist */
+	if (currentReferrer && pluginsOn
+	    && (mt = findMimeByURL(currentReferrer)) && mt->stream) {
+		cf->mt = mt;
+		goto mimestream;
+	}
+
 	cbd.buffer = &serverData;
 	cbd.length = &serverDataLen;
 	cbd.down_state = 0;
@@ -850,6 +962,7 @@ mimestream:
 	urlSanitize(url, post, f_encoded);
 
 	if (post) {
+		post_request = true;
 		post++;
 
 		if (strncmp(post, "`mfd~", 5)) ;	/* No need to do anything, not multipart. */
@@ -953,18 +1066,70 @@ mimestream:
 	still_fetching = true;
 	serverData = initString(&serverDataLen);
 
+	if (!post_request && presentInCache(urlcopy)) {
+		head_request = true;
+		curl_easy_setopt(h, CURLOPT_NOBODY, 1l);
+	}
+
 	while (still_fetching == true) {
 		char *redir = NULL;
 		cbd.length = &serverDataLen;
+// An earlier 302 redirection could set content type = application/binary,
+// which in turn sets state = 1, which is ignored since 302 takes precedence.
+// So state might still be 1, set it back to 0.
+		cbd.down_state = 0;
+
+// recheck suffix after a redirect
+		if (redirect_count &&
+		    (cf->mt = findMimeByURL(urlcopy)) && pluginsOn
+		    && cf->mt->stream) {
+			curl_easy_cleanup(h);
+			goto mimestream;
+		}
 
 perform:
-		curlret = fetch_internet(h, true);
+		is_http = true;
+		ht_cacheable = true;
+		curlret = fetch_internet(h);
 
 		if (cbd.down_state == 6) {
 			curl_easy_cleanup(h);
 			goto mimestream;
 		}
 
+/*********************************************************************
+If the desired file is in cache for some reason, and we issued the head request,
+and it is application, or some such that triggers a download, then state = 1,
+but no data is forthcoming, and the user was never asked if he wants
+to download, so state is still 1.
+So ask, and then look at state.
+If state is nonzero, sorry, I'm not taking the file from cache,
+not yet, just because it's a whole bunch of new code.
+We don't generally store our downloaded files in cache anyways,
+they go where they go, so this doesn't come up very often.
+*********************************************************************/
+
+		if (head_request) {
+			if (cbd.down_state == 1) {
+				setup_download(&cbd);
+/* now we have our answer */
+			}
+
+			if (cbd.down_state != 0) {
+				curl_easy_setopt(h, CURLOPT_NOBODY, 0l);
+				head_request = false;
+				debugPrint(3, "switch to get for download %d",
+					   cbd.down_state);
+			}
+
+			if (cbd.down_state == 2) {
+				curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE,
+						  &ht_code);
+				if (ht_code == 200)
+					goto perform;
+			}
+		}
+
 		if (cbd.down_state == 5) {
 /* user has directed a download of this file in the background. */
 			background_download(&cbd);
@@ -1005,32 +1170,32 @@ perform:
 
 		if (curlret != CURLE_OK)
 			goto curl_fail;
-		curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE, &hcode);
+		curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE, &ht_code);
 		if (curlret != CURLE_OK)
 			goto curl_fail;
 
-		debugPrint(3, "http code %ld %s", hcode, herror);
+		debugPrint(3, "http code %ld", ht_code);
 
 /* refresh header is an alternate form of redirection */
 		if (newlocation && newloc_d >= 0) {
-			if (!refreshDelay(newloc_d, newlocation)) {
+			if (shortRefreshDelay()) {
+				ht_code = 302;
+			} else {
 				nzFree(newlocation);
 				newlocation = 0;
-			} else {
-				hcode = 302;
 			}
 		}
 
 		if (allowRedirection &&
-		    (hcode >= 301 && hcode <= 303 ||
-		     hcode >= 307 && hcode <= 308)) {
+		    (ht_code >= 301 && ht_code <= 303 ||
+		     ht_code >= 307 && ht_code <= 308)) {
 			redir = newlocation;
 			if (redir)
 				redir = resolveURL(urlcopy, redir);
 			still_fetching = false;
 			if (redir == NULL) {
 				/* Redirected, but we don't know where to go. */
-				i_printf(MSG_RedirectNoURL, hcode);
+				i_printf(MSG_RedirectNoURL, ht_code);
 				transfer_status = true;
 			} else if (redirect_count >= 10) {
 				i_puts(MSG_RedirectMany);
@@ -1044,8 +1209,10 @@ perform:
 
 /* Convert POST request to GET request after redirection. */
 /* This should only be done for 301 through 303 */
-				if (hcode < 307)
+				if (ht_code < 307) {
 					curl_easy_setopt(h, CURLOPT_HTTPGET, 1);
+					post_request = false;
+				}
 /* I think there is more work to do for 307 308,
  * pasting the prior post string onto the new URL. Not sure about this. */
 
@@ -1061,6 +1228,11 @@ perform:
 				if (curlret != CURLE_OK)
 					goto curl_fail;
 
+				if (!post_request && presentInCache(urlcopy)) {
+					head_request = true;
+					curl_easy_setopt(h, CURLOPT_NOBODY, 1l);
+				}
+
 				nzFree(serverData);
 				serverData = emptyString;
 				serverDataLen = 0;
@@ -1071,7 +1243,7 @@ perform:
 			}
 		}
 
-		else if (hcode == 401 && !proceed_unauthenticated) {
+		else if (ht_code == 401 && !proceed_unauthenticated) {
 			bool got_creds;
 			i_printf(MSG_AuthRequired, urlcopy);
 			nl();
@@ -1087,8 +1259,31 @@ perform:
 				proceed_unauthenticated = true;
 			}
 		} else {	/* not redirect, not 401 */
-			still_fetching = false;
-			transfer_status = true;
+			if (head_request) {
+				if (fetchCache
+				    (urlcopy, ht_etag, ht_modtime, &cacheData,
+				     &cacheDataLen)) {
+					nzFree(serverData);
+					serverData = cacheData;
+					serverDataLen = cacheDataLen;
+					still_fetching = false;
+					transfer_status = true;
+				} else {
+/* Back through the loop,
+ * now doing GET rather than HEAD. */
+					curl_easy_setopt(h, CURLOPT_NOBODY, 0l);
+					head_request = false;
+					--redirect_count;
+				}
+			} else {
+				if (ht_code == 200 && ht_cacheable &&
+				    (ht_modtime || ht_etag) &&
+				    cbd.down_state == 0)
+					storeCache(urlcopy, ht_etag, ht_modtime,
+						   serverData, serverDataLen);
+				still_fetching = false;
+				transfer_status = true;
+			}
 		}
 	}
 
@@ -1110,11 +1305,11 @@ curl_fail:
 		serverDataLen = 0;
 		nzFree(urlcopy);	/* Free it on transfer failure. */
 	} else {
-		if (hcode != 200 && hcode != 201 &&
+		if (ht_code != 200 && ht_code != 201 &&
 		    (webpage || debugLevel >= 2) ||
-		    hcode == 201 && debugLevel >= 3)
+		    ht_code == 201 && debugLevel >= 3)
 			i_printf(MSG_HTTPError,
-				 hcode, message_for_response_code(hcode));
+				 ht_code, message_for_response_code(ht_code));
 		if (name_changed)
 			changeFileName = urlcopy;
 		else
@@ -1123,9 +1318,16 @@ curl_fail:
 
 	nzFree(postb);
 
+/* see if http header has set the filename */
+	if (ht_cdfn) {
+		nzFree(changeFileName);
+		changeFileName = ht_cdfn;
+		ht_cdfn = NULL;
+	}
+
 /* Check for plugin to run here */
-	if (transfer_status && hcode == 200 && cw->mt && pluginsOn &&
-	    !cw->mt->stream && !cw->mt->outtype && cw->mt->program) {
+	if (transfer_status && ht_code == 200 && cf->mt && pluginsOn &&
+	    !cf->mt->stream && !cf->mt->outtype && cf->mt->program) {
 		bool rc = playServerData();
 		nzFree(serverData);
 		serverData = NULL;
@@ -1352,7 +1554,7 @@ void ebcurl_setError(CURLcode curlret, const char *url)
 		break;
 
 	case CURLE_SSL_CONNECT_ERROR:
-		setError(MSG_SSLConnectError, errorText);
+		setError(MSG_SSLConnectError, ht_error);
 		break;
 
 	case CURLE_LOGIN_DENIED:
@@ -1378,6 +1580,7 @@ static bool ftpConnect(const char *url, const char *user, const char *pass,
 	char creds_buf[MAXUSERPASS * 2 + 1];
 	size_t creds_len = 0;
 
+	is_http = false;
 	protLength = strchr(url, ':') - url + 3;
 /* scp is somewhat unique among the protocols handled here */
 	is_scp = memEqualCI(url, "scp", 3);
@@ -1426,7 +1629,7 @@ static bool ftpConnect(const char *url, const char *user, const char *pass,
 	cbd.length = &serverDataLen;
 
 perform:
-	curlret = fetch_internet(h, false);
+	curlret = fetch_internet(h);
 
 	if (cbd.down_state == 5) {
 /* user has directed a download of this file in the background. */
@@ -1481,7 +1684,7 @@ perform:
 			if (curlret != CURLE_OK)
 				goto ftp_transfer_fail;
 
-			curlret = fetch_internet(h, false);
+			curlret = fetch_internet(h);
 			if (curlret != CURLE_OK)
 				transfer_success = false;
 			else {
@@ -1598,8 +1801,8 @@ static CURL *http_curl_init(struct eb_curl_callback_data *cbd)
 			 CURLAUTH_BASIC | CURLAUTH_DIGEST | CURLAUTH_NTLM);
 
 /* The next few setopt calls could allocate or perform file I/O. */
-	errorText[0] = '\0';
-	curl_init_status = curl_easy_setopt(h, CURLOPT_ERRORBUFFER, errorText);
+	ht_error[0] = '\0';
+	curl_init_status = curl_easy_setopt(h, CURLOPT_ERRORBUFFER, ht_error);
 	if (curl_init_status != CURLE_OK)
 		goto libcurl_init_fail;
 	curl_init_status = curl_easy_setopt(h, CURLOPT_ENCODING, "");
@@ -1801,7 +2004,7 @@ curl_header_callback(char *header_line, size_t size, size_t nmemb,
 
 	scan_http_headers(true);
 	if (down_permitted && data->down_state == 0) {
-		if (cw->mt && cw->mt->stream && pluginsOn) {
+		if (cf->mt && cf->mt->stream && pluginsOn) {
 /* I don't think this ever happens, since streams are indicated by the protocol,
  * and we wouldn't even get here, but just in case -
  * stop the download and set the flag so we can pass this url
@@ -1809,10 +2012,13 @@ curl_header_callback(char *header_line, size_t size, size_t nmemb,
 			data->down_state = 6;
 			return -1;
 		}
-		if (hct[0] && !memEqualCI(hct, "text/", 5) &&
-		    (!pluginsOn || !cw->mt || cw->mt->download)) {
+		if (ht_content[0] && !memEqualCI(ht_content, "text/", 5) &&
+		    !memEqualCI(ht_content, "application/xhtml+xml", 21) &&
+		    (!pluginsOn || !cf->mt || cf->mt->download)) {
 			data->down_state = 1;
 			down_msg = MSG_Down;
+			debugPrint(3, "potential download based on type %s",
+				   ht_content);
 		}
 	}
 
@@ -1864,6 +2070,7 @@ static void setup_download(struct eb_curl_callback_data *data)
 {
 	const char *filepart;
 	const char *answer;
+	const struct MIMETYPE *mt;
 
 /* if not run from a terminal then just return. */
 	if (!isInteractive) {
@@ -1871,6 +2078,13 @@ static void setup_download(struct eb_curl_callback_data *data)
 		return;
 	}
 
+/* If file is changed to a playlist or some such, just return */
+	if (ht_cdfn && (mt = findMimeByURL(ht_cdfn)) && mt->stream) {
+		debugPrint(3, "download aborted due to stream plugin");
+		data->down_state = 0;
+		return;
+	}
+
 	filepart = getFileURL(urlcopy, true);
 top:
 	answer = getFileName(down_msg, filepart, false, true);
@@ -1954,7 +2168,7 @@ static void background_download(struct eb_curl_callback_data *data)
 		job->file2 = data->down_file2 - data->down_file;
 // round file size up to the nearest chunk.
 // This will come out 0 only if the true size is 0.
-		job->fsize = ((hcl + (CHUNKSIZE - 1)) / CHUNKSIZE);
+		job->fsize = ((ht_length + (CHUNKSIZE - 1)) / CHUNKSIZE);
 		addToListBack(&down_jobs, job);
 
 		return;
@@ -2003,7 +2217,7 @@ int bg_jobs(bool iponly)
 		}
 		printf("%s", j->file + j->file2);
 		if (j->fsize)
-			printf(" %d/%d",
+			printf(" %d/%lu",
 			       (fileSizeByName(j->file) / CHUNKSIZE), j->fsize);
 		nl();
 	}
@@ -2108,10 +2322,12 @@ CURLcode setCurlURL(CURL * h, const char *url)
 	else
 		debugPrint(3, "proxy %s", proxy);
 	host = getHostURL(url);
+	if (!host)		// should never happen
+		return CURLE_URL_MALFORMAT;
 	verify = mustVerifyHost(host);
 	curl_easy_setopt(h, CURLOPT_PROXY, proxy);
 	curl_easy_setopt(h, CURLOPT_SSL_VERIFYPEER, verify);
-	curl_easy_setopt(h, CURLOPT_SSL_VERIFYHOST, verify);
+	curl_easy_setopt(h, CURLOPT_SSL_VERIFYHOST, (verify ? 2 : 0));
 	return curl_easy_setopt(h, CURLOPT_URL, url);
 }				/* setCurlURL */
 
@@ -2176,3 +2392,287 @@ const char *findProxyForURL(const char *url)
 {
 	return findProxyInternal(getProtURL(url), getHostURL(url));
 }				/* findProxyForURL */
+
+/* expand a frame inline.
+ * Pass a range of lines; you can expand all the frames in one go.
+ * Return false if there is a problem fetching a web page,
+ * or if none of the lines are frames. */
+static int frameExpandLine(int lineNumber);
+static int frameContractLine(int lineNumber);
+static const char *stringInBufLine(const char *s, const char *t);
+bool frameExpand(bool expand, int ln1, int ln2)
+{
+	int ln;			/* line number */
+	int problem = 0, p;
+	bool something_worked = false;
+
+	for (ln = ln1; ln <= ln2; ++ln) {
+		if (expand)
+			p = frameExpandLine(ln);
+		else
+			p = frameContractLine(ln);
+		if (p > problem)
+			problem = p;
+		if (p == 0)
+			something_worked = true;
+	}
+
+	if (something_worked && problem < 3)
+		problem = 0;
+	if (problem == 1)
+		setError(expand ? MSG_NoFrame1 : MSG_NoFrame2);
+	if (problem == 2)
+		setError(MSG_FrameNoURL);
+	return (problem == 0);
+}				/* frameExpand */
+
+/* Problems: 0, frame expanded successfully.
+ 1 line is not a frame.
+ 2 frame doesn't have a valid url.
+ 3 Problem fetching the rul or rendering the page.  */
+static int frameExpandLine(int ln)
+{
+	pst line;
+	int tagno, start;
+	const char *s;
+	struct htmlTag *t;
+	struct ebFrame *save_cf, *last_f;
+
+	line = fetchLine(ln, -1);
+	s = stringInBufLine(line, "Frame ");
+	if (!s)
+		return 1;
+	if ((s = strchr(s, InternalCodeChar)) == NULL)
+		return 2;
+	tagno = strtol(s + 1, (char **)&s, 10);
+	if (tagno < 0 || tagno >= cw->numTags || *s != '{')
+		return 2;
+	t = tagList[tagno];
+	if (t->action != TAGACT_FRAME)
+		return 1;
+
+/* the easy case is if it's already been expanded before, we just unhide it. */
+	if (t->f1) {
+		t->contracted = false;
+		return 0;
+	}
+
+	s = t->href;
+	if (!s)
+		return 2;
+
+	save_cf = cf;
+/* have to push a new frame before we read the web page */
+	for (last_f = &(cw->f0); last_f->next; last_f = last_f->next) ;
+	last_f->next = cf = allocZeroMem(sizeof(struct ebFrame));
+	cf->owner = cw;
+	cf->frametag = t;
+	debugPrint(2, "fetch frame %s", s);
+	if (!readFileArgv(s)) {
+/* serverData was never set, or was freed do to some other error. */
+/* We just need to pop the frame and return. */
+		fileSize = -1;	/* don't print 0 */
+		nzFree(cf->fileName);
+		free(cf);
+		last_f->next = 0;
+		cf = save_cf;
+		return 3;
+	}
+
+/*********************************************************************
+readFile could return success and yet serverData is null.
+This happens if httpConnect did something other than fetching data,
+like playing a stream. Does that happen, even in a frame?
+It can, if the frame is a youtube video, which is not unusual at all.
+So check for serverData null here. Once again we pop the frame.
+*********************************************************************/
+
+	if (serverData == NULL) {
+		nzFree(cf->fileName);
+		free(cf);
+		last_f->next = 0;
+		cf = save_cf;
+		fileSize = -1;
+		return 0;
+	}
+
+	if (changeFileName) {
+		nzFree(cf->fileName);
+		cf->fileName = changeFileName;
+		cf->f_encoded = true;
+		changeFileName = 0;
+	} else {
+		cf->fileName = cloneString(s);
+	}
+
+/* don't print the size of what we just fetched */
+	fileSize = -1;
+
+/* If we got some data it has to be html.
+ * I should check for that, something like htmlTest in html.c,
+ * but I'm too lazy to do that right now, so I'll just assume it's good. */
+
+	cf->hbase = cloneString(cf->fileName);
+	browseLocal = !isURL(cf->fileName);
+	prepareForBrowse(serverData, serverDataLen);
+	if (javaOK(cf->fileName))
+		createJavaContext();
+	nzFree(newlocation);	/* should already be 0 */
+	newlocation = 0;
+
+	start = cw->numTags;
+/* call the tidy parser to build the html nodes */
+	html2nodes(serverData, true);
+	nzFree(serverData);	/* don't need it any more */
+	htmlGenerated = false;
+	htmlNodesIntoTree(start, t);
+	prerender(0);
+	if (isJSAlive) {
+		decorate(0);
+		set_basehref(cf->hbase);
+		runScriptsPending();
+		runOnload();
+		runScriptsPending();
+		set_property_string(cf->docobj, "readyState", "complete");
+	}
+
+	if (cf->fileName) {
+		int j = strlen(cf->fileName);
+		cf->fileName = reallocMem(cf->fileName, j + 8);
+		strcat(cf->fileName, ".browse");
+	}
+
+	t->f1 = cf;
+	cf = save_cf;
+	return 0;
+}				/* frameExpandLine */
+
+static int frameContractLine(int ln)
+{
+	struct htmlTag *t = line2frame(ln);
+	if (!t)
+		return 1;
+	t->contracted = true;
+	return 0;
+}				/* frameContractLine */
+
+struct htmlTag *line2frame(int ln)
+{
+	const char *line;
+	int n, opentag = 0, ln1 = ln;
+	const char *s;
+
+	for (; ln; --ln) {
+		line = (char *)fetchLine(ln, -1);
+		if (!opentag && ln < ln1
+		    && (s = stringInBufLine(line, "*--`\n"))) {
+			for (--s; s > line && *s != InternalCodeChar; --s) ;
+			if (*s == InternalCodeChar)
+				opentag = atoi(s + 1);
+			continue;
+		}
+		s = stringInBufLine(line, "*`--\n");
+		if (!s)
+			continue;
+		for (--s; s > line && *s != InternalCodeChar; --s) ;
+		if (*s != InternalCodeChar)
+			continue;
+		n = atoi(s + 1);
+		if (!opentag)
+			return tagList[n];
+		if (n == opentag)
+			opentag = 0;
+	}
+
+	return 0;
+}				/* line2frame */
+
+/* a text line in the buffer isn't a string; you can't use strstr */
+static const char *stringInBufLine(const char *s, const char *t)
+{
+	int n = strlen(t);
+	for (; *s != '\n'; ++s) {
+		if (!strncmp(s, t, n))
+			return s;
+	}
+	return 0;
+}				/* stringInBufLine */
+
+bool reexpandFrame(void)
+{
+	int j, start;
+	struct htmlTag *t, *frametag;
+
+/* cut the children off from the frame tag */
+	cf = newloc_f;
+	frametag = cf->frametag;
+	for (t = frametag->firstchild; t; t = t->sibling) {
+		t->deleted = true;
+		t->step = 100;
+		t->parent = 0;
+	}
+	frametag->firstchild = 0;
+
+	delTimers(cf);
+	delInputChanges(cf);
+	freeJavaContext(cf);
+	nzFree(cf->dw);
+	cf->dw = 0;
+	nzFree(cf->hbase);
+	cf->hbase = 0;
+	nzFree(cf->fileName);
+	cf->fileName = newlocation;
+	newlocation = 0;
+	cf->f_encoded = false;
+	nzFree(cf->firstURL);
+	cf->firstURL = 0;
+
+	if (!readFileArgv(cf->fileName)) {
+/* serverData was never set, or was freed do to some other error. */
+		fileSize = -1;	/* don't print 0 */
+		return false;
+	}
+
+	if (serverData == NULL) {
+/* frame replaced itself with a playable stream, what to do? */
+		fileSize = -1;
+		return true;
+	}
+
+	if (changeFileName) {
+		nzFree(cf->fileName);
+		cf->fileName = changeFileName;
+		cf->f_encoded = true;
+		changeFileName = 0;
+	}
+
+/* don't print the size of what we just fetched */
+	fileSize = -1;
+
+	cf->hbase = cloneString(cf->fileName);
+	browseLocal = !isURL(cf->fileName);
+	prepareForBrowse(serverData, serverDataLen);
+	if (javaOK(cf->fileName))
+		createJavaContext();
+	start = cw->numTags;
+/* call the tidy parser to build the html nodes */
+	html2nodes(serverData, true);
+	nzFree(serverData);	/* don't need it any more */
+	htmlGenerated = false;
+	htmlNodesIntoTree(start, frametag);
+	prerender(0);
+	if (isJSAlive) {
+		decorate(0);
+		set_basehref(cf->hbase);
+		runScriptsPending();
+		runOnload();
+		runScriptsPending();
+		set_property_string(cf->docobj, "readyState", "complete");
+	}
+
+	j = strlen(cf->fileName);
+	cf->fileName = reallocMem(cf->fileName, j + 8);
+	strcat(cf->fileName, ".browse");
+
+	return true;
+}				/* reexpandFrame */
