@@ -56,6 +56,11 @@ static void usage(void)
 /* arguments, as indicated by the above */
 static int pipe_in, pipe_out;
 
+/* Here is an instance of the edbrowse window that exists for parsing
+ * html from inside javascript. It belongs to edbrowse-js,
+ * and isn't associated with a particular buffer on the edbrowse side. */
+static struct ebWindow in_js_cw;
+
 static void js_start(void);
 static void readMessage(void);
 static void processMessage(void);
@@ -68,10 +73,10 @@ static JSObject *docobj;	/* document object */
 
 static void cwSetup(void)
 {
-	in_js_cw.winobj = winobj;
-	in_js_cw.docobj = docobj;
-	in_js_cw.hbase = get_property_string_nat(docobj, "base$href");
-	in_js_cw.baseset = true;
+	cf->winobj = winobj;
+	cf->docobj = docobj;
+	cf->hbase = get_property_string_nat(docobj, "base$href");
+	cf->baseset = true;
 }				/* cwSetup */
 
 static void cwBringdown(void)
@@ -79,8 +84,8 @@ static void cwBringdown(void)
 	freeTags(cw);
 	nzFree(cw->ft);		/* title could have been set by prerender */
 	cw->ft = 0;
-	nzFree(cw->hbase);
-	cw->hbase = 0;
+	nzFree(cf->hbase);
+	cf->hbase = 0;
 }				/* cwBringdown */
 
 static struct EJ_MSG head;
@@ -124,14 +129,20 @@ int js_main(int argc, char **argv)
 		usage();
 
 	readConfigFile();
+	setupEdbrowseCache();
 	eb_curl_global_init();
 	cookiesFromJar();
 	pluginsOn = false;
+	sendReferrer = false;
+	cw = &in_js_cw;
+	cf = &(cw->f0);
+	cf->owner = cw;
 
 	js_start();
 
 /* edbrowse catches interrupt, this process ignores it. */
 /* Use quit to terminate, or kill from another console. */
+/* If edbrowse quits then this process also quits via broken pipe. */
 	signal(SIGINT, SIG_IGN);
 
 	effects = initString(&eff_l);
@@ -173,6 +184,15 @@ int js_main(int argc, char **argv)
 		if (head.cmd == EJ_CMD_VARUPDATE) {
 			if (head.lineno == 1)
 				allowXHR = head.n;
+			if (head.lineno == 2)
+				debugLevel = head.n;
+			if (head.lineno == 3)
+				verifyCertificates = head.n;
+			if (head.lineno == 4) {
+				char *t = userAgents[head.n];
+				if (t)
+					currentAgent = t;
+			}
 			head.n = head.proplength = 0;
 //                      no acknowledgement needed
 //                      writeHeader();
@@ -185,16 +205,22 @@ int js_main(int argc, char **argv)
 }				/* js_main */
 
 /* read from and write to edbrowse */
+
 static void readFromEb(void *data_p, int n)
 {
 	int rc;
+	unsigned char *bytes_p = (unsigned char *)data_p;
 	if (n == 0)
 		return;
-	rc = read(pipe_in, data_p, n);
-	if (rc == n)
-		return;
+	while (n > 0) {
+		rc = read(pipe_in, bytes_p, n);
+		if (rc <= 0) {
 /* Oops - can't read from the process any more */
-	exit(2);
+			exit(2);
+		}
+		n -= rc;
+		bytes_p += rc;
+	}
 }				/* readFromEb */
 
 static void writeToEb(const void *data_p, int n)
@@ -309,13 +335,14 @@ static void js_start(void)
 	exit(4);
 }				/* js_start */
 
-static void misconfigure(void)
+static void misconfigure(int n)
 {
 /* there may already be a larger error */
 	if (head.highstat > EJ_HIGH_CX_FAIL)
 		return;
 	head.highstat = EJ_HIGH_CX_FAIL;
 	head.lowstat = EJ_LOW_VARS;
+	head.lineno = n;
 }				/* misconfigure */
 
 static void
@@ -384,7 +411,7 @@ static char *JS_c_str(js::HandleString str)
 	size_t result =
 	    JS_EncodeStringToBuffer(jcx, str, buffer, encodedLength);
 	if (result == (size_t) - 1) {
-		misconfigure();
+		misconfigure(2);
 		buffer[0] = 0;
 	}
 	return buffer;
@@ -484,7 +511,7 @@ static JSBool c##_ctor(JSContext * cx, unsigned int argc, jsval * vp) \
 	JS::RootedObject newobj(cx, \
 		JS_NewObjectForConstructor(cx, &c##_class, &callee_val)); \
 	if (newobj == NULL) { \
-		misconfigure(); \
+		misconfigure(1); \
 		return JS_FALSE; \
 	} \
 	args.rval().set(OBJECT_TO_JSVAL(newobj)); \
@@ -520,7 +547,7 @@ static JSBool window_ctor(JSContext * cx, unsigned int argc, jsval * vp)
 				JS_NewObjectForConstructor(cx, &window_class,
 							   &callee_val));
 	if (newwin == NULL) {
-		misconfigure();
+		misconfigure(3);
 		return JS_FALSE;
 	}
 	if (args.length() > 0 && (str = JS_ValueToString(cx, args[0])))
@@ -594,7 +621,7 @@ static JSBool option_ctor(JSContext * cx, unsigned int argc, jsval * vp)
 				JS_NewObjectForConstructor(cx, &option_class,
 							   &callee_val));
 	if (newopt == NULL) {
-		misconfigure();
+		misconfigure(4);
 		return JS_FALSE;
 	}
 	if (args.length() > 0 && (str = JS_ValueToString(cx, args[0]))) {
@@ -628,7 +655,7 @@ static JSBool textnode_ctor(JSContext * cx, unsigned int argc, jsval * vp)
 				 JS_NewObjectForConstructor(cx, &textnode_class,
 							    &callee_val));
 	if (newtext == NULL) {
-		misconfigure();
+		misconfigure(5);
 		return JS_FALSE;
 	}
 	if (args.length() > 0 && (str = JS_ValueToString(cx, args[0])))
@@ -658,7 +685,7 @@ static JSBool url_ctor(JSContext * cx, unsigned int argc, jsval * vp)
 						       &callee_val));
 	if (uo == NULL) {
 abort:
-		misconfigure();
+		misconfigure(6);
 		return JS_FALSE;
 	}
 
@@ -705,6 +732,7 @@ static JSBool
 setter_loc(JSContext * cx, JS::HandleObject uo, JS::Handle < jsid > id,
 	   JSBool strict, JS::MutableHandle < JS::Value > vp)
 {
+	debugPrint(5, "setter location");
 	const char *s = stringize(vp);
 	if (!s) {
 		JS_ReportError(jcx,
@@ -715,6 +743,7 @@ setter_loc(JSContext * cx, JS::HandleObject uo, JS::Handle < jsid > id,
 		effectChar('\n');
 		endeffect();
 	}
+	debugPrint(5, "return");
 	return JS_FALSE;
 }				/* setter_loc */
 
@@ -729,16 +758,17 @@ setter_loc_hrefval(JSContext * cx, JS::HandleObject uo,
 	const char *url = 0;
 	if (setter_suspend)
 		return JS_TRUE;
+	debugPrint(5, "setter location href");
 	url = stringize(vp);
-	if (!url)
-		return JS_TRUE;
-	if (iswindocloc(uo)) {
+	if (url && iswindocloc(uo)) {
 		effectString("n{r");	// }
 		effectString(url);
 		effectChar('\n');
 		endeffect();
+		debugPrint(5, "return abort");
 		return JS_FALSE;
 	}
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* setter_loc_hrefval */
 
@@ -869,6 +899,7 @@ setter_value(JSContext * cx, JS::HandleObject obj,
 	const char *val;
 	if (setter_suspend)
 		return JS_TRUE;
+	debugPrint(5, "setter value");
 	val = stringize(vp);
 	if (!val) {
 		JS_ReportError(jcx,
@@ -880,6 +911,7 @@ setter_value(JSContext * cx, JS::HandleObject obj,
 		effectString(val);
 		endeffect();
 	}
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* setter_value */
 
@@ -892,6 +924,8 @@ setter_innerHTML(JSContext * cx, JS::HandleObject obj,
 	if (!s)
 		return JS_TRUE;
 
+	debugPrint(5, "setter innerHTML");
+
 /* lop off the preexisting children */
 	JS::RootedObject children(jcx);
 	children = get_property_object1(obj, "childNodes");
@@ -914,6 +948,7 @@ setter_innerHTML(JSContext * cx, JS::HandleObject obj,
 	packDecoration();
 	cwBringdown();
 	endeffect();
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* setter_innerHTML */
 
@@ -922,6 +957,7 @@ setter_innerText(JSContext * cx, JS::HandleObject obj,
 		 JS::Handle < jsid > id, JSBool strict,
 		 JS::MutableHandle < jsval > vp)
 {
+	debugPrint(5, "setter innerText");
 	const char *s = stringize(vp);
 	if (!s)
 		s = emptyString;
@@ -930,6 +966,7 @@ setter_innerText(JSContext * cx, JS::HandleObject obj,
 	effectChar('|');
 	effectString(s);
 	endeffect();
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* setter_innerText */
 
@@ -1010,6 +1047,8 @@ setter_cookie(JSContext * cx, JS::HandleObject obj,
 	if (setter_suspend)
 		return JS_TRUE;
 
+	debugPrint(5, "setter cookie");
+
 /* grab the existing document.cookie string, is this reentrant, is this ok? */
 	original = get_property_string1(obj, "cookie");
 	if (!original)		/* should never happen */
@@ -1031,6 +1070,7 @@ setter_cookie(JSContext * cx, JS::HandleObject obj,
 		vp.set(v);
 	}
 	nzFree(cookieCopy);
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* setter_cookie */
 
@@ -1043,7 +1083,9 @@ setter_domain(JSContext * cx, JS::HandleObject obj,
 	const char *val;
 	if (setter_suspend)
 		return JS_TRUE;
+	debugPrint(5, "setter domain");
 	val = stringize(vp);
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* setter_domain */
 
@@ -1064,7 +1106,7 @@ set_property_string1(js::HandleObject parent, const char *name,
 	JS_HasProperty(jcx, parent, name, &found);
 	if (found) {
 		if (JS_SetProperty(jcx, parent, name, v.address()) == JS_FALSE)
-			misconfigure();
+			misconfigure(7);
 		return;
 	}
 
@@ -1081,7 +1123,7 @@ set_property_string1(js::HandleObject parent, const char *name,
 
 	if (JS_DefineProperty
 	    (jcx, parent, name, v, my_getter, my_setter, PROP_STD) == JS_FALSE)
-		misconfigure();
+		misconfigure(8);
 }				/* set_property_string1 */
 
 int set_property_string_nat(jsobjtype parent, const char *name,
@@ -1103,13 +1145,13 @@ static void set_property_bool1(js::HandleObject parent, const char *name,
 	JS_HasProperty(jcx, parent, name, &found);
 	if (found) {
 		if (JS_SetProperty(jcx, parent, name, v.address()) == JS_FALSE)
-			misconfigure();
+			misconfigure(9);
 		return;
 	}
 
 	if (JS_DefineProperty(jcx, parent, name, v, NULL, NULL, PROP_STD) ==
 	    JS_FALSE)
-		misconfigure();
+		misconfigure(10);
 }				/* set_property_bool1 */
 
 int set_property_bool_nat(jsobjtype parent, const char *name, bool n)
@@ -1130,13 +1172,13 @@ static void set_property_number1(js::HandleObject parent, const char *name,
 	JS_HasProperty(jcx, parent, name, &found);
 	if (found) {
 		if (JS_SetProperty(jcx, parent, name, v.address()) == JS_FALSE)
-			misconfigure();
+			misconfigure(11);
 		return;
 	}
 
 	if (JS_DefineProperty(jcx, parent, name, v, NULL, NULL, PROP_STD) ==
 	    JS_FALSE)
-		misconfigure();
+		misconfigure(12);
 }				/* set_property_number1 */
 
 int set_property_number_nat(jsobjtype parent, const char *name, int n)
@@ -1157,13 +1199,13 @@ static void set_property_float1(js::HandleObject parent, const char *name,
 	JS_HasProperty(jcx, parent, name, &found);
 	if (found) {
 		if (JS_SetProperty(jcx, parent, name, v.address()) == JS_FALSE)
-			misconfigure();
+			misconfigure(13);
 		return;
 	}
 
 	if (JS_DefineProperty(jcx, parent, name, v, NULL, NULL, PROP_STD) ==
 	    JS_FALSE)
-		misconfigure();
+		misconfigure(14);
 }				/* set_property_float1 */
 
 int set_property_float_nat(jsobjtype parent, const char *name, double n)
@@ -1186,7 +1228,7 @@ set_property_object1(js::HandleObject parent, const char *name,
 	JS_HasProperty(jcx, parent, name, &found);
 	if (found) {
 		if (JS_SetProperty(jcx, parent, name, v.address()) == JS_FALSE)
-			misconfigure();
+			misconfigure(15);
 		return;
 	}
 
@@ -1195,7 +1237,7 @@ set_property_object1(js::HandleObject parent, const char *name,
 
 	if (JS_DefineProperty(jcx, parent, name, v, NULL, my_setter, PROP_STD)
 	    == JS_FALSE)
-		misconfigure();
+		misconfigure(16);
 }				/* set_property_object1 */
 
 int set_property_object_nat(jsobjtype parent, const char *name, jsobjtype child)
@@ -1267,26 +1309,6 @@ void run_function_onearg_nat(jsobjtype parent, const char *name,
 	run_function_onearg1(p, name, c);
 }				/* run_function_onearg_nat */
 
-#if 0
-/* Not clear that setAttribute needs any side effects, or needs to be native. */
-static JSBool setAttribute(JSContext * cx, unsigned int argc, jsval * vp)
-{
-	JS::RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
-	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
-	js::RootedValue v1(cx), v2(cx);
-	if (args.length() != 2 || !JSVAL_IS_STRING(args[0])) {
-		JS_ReportError(cx, "unexpected arguments to setAttribute()");
-	} else {
-		v1 = args[0];
-		v2 = args[1];
-		const char *prop = stringize(v1);
-		JS_DefineProperty(cx, thisobj, prop, v2, NULL, NULL, PROP_STD);
-	}
-	args.rval().set(JSVAL_VOID);
-	return JS_TRUE;
-}				/* setAttribute */
-#endif
-
 static void embedNodeName(JS::HandleObject obj)
 {
 	const char *nodeName;
@@ -1317,20 +1339,20 @@ static JSBool appendChild0(bool side, JSContext * cx, unsigned int argc,
 
 	JS::RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
 	if (JS_GetProperty(cx, thisobj, "childNodes", v.address()) == JS_FALSE)
-		return JS_TRUE;	/* no such array */
+		return JS_FALSE;	/* error */
 	JS::RootedObject elar(cx, JSVAL_TO_OBJECT(v));
 	if (elar == NULL) {
-		misconfigure();
+		misconfigure(17);
 		return JS_FALSE;
 	}
 	if (JS_GetArrayLength(cx, elar, &length) == JS_FALSE) {
-		misconfigure();
+		misconfigure(18);
 		return JS_FALSE;
 	}
 // see if it's already there.
 	for (i = 0; i < length; ++i) {
 		if (JS_GetElement(cx, elar, i, v.address()) == JS_FALSE) {
-			misconfigure();
+			misconfigure(19);
 			return JS_FALSE;
 		}
 		if (!v.isObject())
@@ -1345,7 +1367,7 @@ static JSBool appendChild0(bool side, JSContext * cx, unsigned int argc,
 // add child to the end
 	if (JS_DefineElement(cx, elar, length,
 			     args[0], NULL, NULL, PROP_STD) == JS_FALSE) {
-		misconfigure();
+		misconfigure(20);
 		return JS_FALSE;
 	}
 	JS_DefineProperty(cx, child, "parentNode", OBJECT_TO_JSVAL(thisobj),
@@ -1369,17 +1391,25 @@ static JSBool appendChild0(bool side, JSContext * cx, unsigned int argc,
 	return JS_TRUE;
 }				/* appendChild0 */
 
-static JSBool appendChild(JSContext * cx, unsigned int argc, jsval * vp)
+static JSBool apch1(JSContext * cx, unsigned int argc, jsval * vp)
 {
-	return appendChild0(true, cx, argc, vp);
-}				/* appendChild */
+	JSBool ret;
+	debugPrint(5, "apch1");
+	ret = appendChild0(false, cx, argc, vp);
+	debugPrint(5, "return");
+	return ret;
+}				/* apch1 */
 
-static JSBool apch(JSContext * cx, unsigned int argc, jsval * vp)
+static JSBool apch2(JSContext * cx, unsigned int argc, jsval * vp)
 {
-	return appendChild0(false, cx, argc, vp);
-}				/* apch */
+	JSBool ret;
+	debugPrint(5, "apch2");
+	ret = appendChild0(true, cx, argc, vp);
+	debugPrint(5, "return");
+	return ret;
+}				/* apch2 */
 
-static JSBool insertBefore(JSContext * cx, unsigned int argc, jsval * vp)
+static JSBool insbf(JSContext * cx, unsigned int argc, jsval * vp)
 {
 	unsigned i, mark, length;
 	js::RootedValue v(cx);
@@ -1389,18 +1419,20 @@ static JSBool insertBefore(JSContext * cx, unsigned int argc, jsval * vp)
 	if (args.length() < 2 || !args[0].isObject() || !args[1].isObject())
 		return JS_TRUE;
 
+	debugPrint(5, "insbf");
+
 	JS::RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
 	JS::RootedObject child(cx, JSVAL_TO_OBJECT(args[0]));
 	JS::RootedObject item(cx, JSVAL_TO_OBJECT(args[1]));
 	if (JS_GetProperty(cx, thisobj, "childNodes", v.address()) == JS_FALSE)
-		return JS_TRUE;	/* no such array */
+		return JS_FALSE;	/* error */
 	JS::RootedObject elar(cx, JSVAL_TO_OBJECT(v));
 	if (elar == NULL) {
-		misconfigure();
+		misconfigure(21);
 		return JS_FALSE;
 	}
 	if (JS_GetArrayLength(cx, elar, &length) == JS_FALSE) {
-		misconfigure();
+		misconfigure(22);
 		return JS_FALSE;
 	}
 
@@ -1408,7 +1440,7 @@ static JSBool insertBefore(JSContext * cx, unsigned int argc, jsval * vp)
 	mark = -1;
 	for (i = 0; i < length; ++i) {
 		if (JS_GetElement(cx, elar, i, v.address()) == JS_FALSE) {
-			misconfigure();
+			misconfigure(23);
 			return JS_FALSE;
 		}
 		if (!v.isObject())
@@ -1418,12 +1450,15 @@ static JSBool insertBefore(JSContext * cx, unsigned int argc, jsval * vp)
 		if (JSVAL_TO_OBJECT(v) == child) {
 // child was already there, what should we do?
 			args.rval().set(args[0]);
+			debugPrint(5, "return");
 			return JS_TRUE;
 		}
 	}
 
-	if (mark < 0)
+	if (mark < 0) {
+		debugPrint(5, "return");
 		return JS_TRUE;
+	}
 
 /* since the item to insert before was found, the call is going to */
 /* succeed, so put the return value here */
@@ -1458,8 +1493,9 @@ static JSBool insertBefore(JSContext * cx, unsigned int argc, jsval * vp)
 	embedNodeName(item);
 	effectChar(' ');
 	endeffect();
+	debugPrint(5, "return");
 	return JS_TRUE;
-}				/* insertBefore */
+}				/* insbf */
 
 static JSBool removeChild(JSContext * cx, unsigned int argc, jsval * vp)
 {
@@ -1471,24 +1507,26 @@ static JSBool removeChild(JSContext * cx, unsigned int argc, jsval * vp)
 	if (args.length() < 1 || !args[0].isObject())
 		return JS_TRUE;
 
+	debugPrint(5, "removeChild");
+
 	JS::RootedObject thisobj(cx, JS_THIS_OBJECT(cx, vp));
 	JS::RootedObject child(cx, JSVAL_TO_OBJECT(args[0]));
 	if (JS_GetProperty(cx, thisobj, "childNodes", v.address()) == JS_FALSE)
-		return JS_TRUE;	/* no such array */
+		return JS_FALSE;	/* error */
 	JS::RootedObject elar(cx, JSVAL_TO_OBJECT(v));
 	if (elar == NULL) {
-		misconfigure();
+		misconfigure(24);
 		return JS_FALSE;
 	}
 	if (JS_GetArrayLength(cx, elar, &length) == JS_FALSE) {
-		misconfigure();
+		misconfigure(25);
 		return JS_FALSE;
 	}
 
 /* child better be somewhere in the array */
 	for (mark = 0; mark < length; ++mark) {
 		if (JS_GetElement(cx, elar, mark, v.address()) == JS_FALSE) {
-			misconfigure();
+			misconfigure(26);
 			return JS_FALSE;
 		}
 		if (!v.isObject())
@@ -1496,6 +1534,7 @@ static JSBool removeChild(JSContext * cx, unsigned int argc, jsval * vp)
 		if (JSVAL_TO_OBJECT(v) == child)
 			goto found;
 	}
+	debugPrint(5, "return");
 	return JS_TRUE;
 
 found:
@@ -1521,11 +1560,13 @@ found:
 	embedNodeName(child);
 	effectString(" 0x0, ");
 	endeffect();
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* removeChild */
 
 static void dwrite1(unsigned int argc, jsval * argv, bool newline)
 {
+	debugPrint(5, "document write");
 	int i, begin;
 	const char *msg;
 	JS::RootedString str(jcx);
@@ -1541,6 +1582,7 @@ static void dwrite1(unsigned int argc, jsval * argv, bool newline)
 	if (newline)
 		effectChar('\n');
 	endeffect();
+	debugPrint(5, "return");
 }				/* dwrite1 */
 
 static JSBool doc_write(JSContext * cx, unsigned int argc, jsval * vp)
@@ -1562,6 +1604,7 @@ static JSBool doc_writeln(JSContext * cx, unsigned int argc, jsval * vp)
 /* this has a native wrapper so we can set innerHTML with a setter */
 static JSBool doc_createElement(JSContext * cx, unsigned int argc, jsval * vp)
 {
+	debugPrint(5, "create element");
 	char run[60];
 	const char *tagname = NULL, *s;
 	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
@@ -1572,6 +1615,7 @@ static JSBool doc_createElement(JSContext * cx, unsigned int argc, jsval * vp)
 fail:
 		cnzFree(tagname);
 		args.rval().set(JSVAL_NULL);
+		debugPrint(5, "return");
 		return JS_TRUE;
 	}
 	for (s = tagname; *s; ++s)
@@ -1598,6 +1642,7 @@ fail:
 
 /* and return the created object */
 	args.rval().set(v);
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* doc_createElement */
 
@@ -1634,6 +1679,7 @@ static void cookieRefresh(void)
 
 static JSBool fetchHTTP(JSContext * cx, unsigned int argc, jsval * vp)
 {
+	debugPrint(5, "fetch http");
 	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
 
 	if (allowXHR) {
@@ -1672,6 +1718,10 @@ static JSBool fetchHTTP(JSContext * cx, unsigned int argc, jsval * vp)
 		httpConnect(curl_incoming_url, false, false, true,
 			    &curl_outgoing_xhrheaders, &curl_outgoing_xhrbody,
 			    &responseLength);
+		if (curl_outgoing_xhrheaders == NULL)
+			curl_outgoing_xhrheaders = emptyString;
+		if (curl_outgoing_xhrbody == NULL)
+			curl_outgoing_xhrbody = emptyString;
 		args.rval().set(STRING_TO_JSVAL
 				(JS_NewStringCopyZ
 				 (cx,
@@ -1699,6 +1749,7 @@ static JSBool fetchHTTP(JSContext * cx, unsigned int argc, jsval * vp)
 		args.rval().set(JS_GetEmptyStringValue(cx));
 	}
 
+	debugPrint(5, "return");
 	return JS_TRUE;
 }
 
@@ -1710,9 +1761,9 @@ static JSFunctionSpec document_methods[] = {
 	JS_FS("write", doc_write, 0, 0),
 	JS_FS("writeln", doc_writeln, 0, 0),
 	JS_FS("createElement", doc_createElement, 0, 0),
-	JS_FS("appendChild", appendChild, 1, 0),
-	JS_FS("apch$", apch, 1, 0),
-	JS_FS("insertBefore", insertBefore, 2, 0),
+	JS_FS("apch1$", apch1, 1, 0),
+	JS_FS("apch2$", apch2, 1, 0),
+	JS_FS("insbf$", insbf, 2, 0),
 	JS_FS("removeChild", removeChild, 1, 0),
 	JS_FS("fetchHTTP", fetchHTTP, 4, 0),
 	JS_FS_END
@@ -1720,23 +1771,27 @@ static JSFunctionSpec document_methods[] = {
 
 static JSBool form_submit(JSContext * cx, unsigned int argc, jsval * vp)
 {
+	debugPrint(5, "form submit");
 	JS::RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
 	effectString("f{s");	// }
 	effectString(pointer2string(obj));
 	endeffect();
 	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
 	args.rval().set(JSVAL_VOID);
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* form_submit */
 
 static JSBool form_reset(JSContext * cx, unsigned int argc, jsval * vp)
 {
+	debugPrint(5, "form reset");
 	JS::RootedObject obj(cx, JS_THIS_OBJECT(cx, vp));
 	effectString("f{r");	// }
 	effectString(pointer2string(obj));
 	endeffect();
 	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
 	args.rval().set(JSVAL_VOID);
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* form_reset */
 
@@ -1771,6 +1826,21 @@ static JSBool win_puts(JSContext * cx, unsigned int argc, jsval * vp)
 	return JS_TRUE;
 }				/* win_puts */
 
+static JSBool win_logputs(JSContext * cx, unsigned int argc, jsval * vp)
+{
+	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
+	const char *msg = NULL;
+	int minlev = JSVAL_TO_INT(args[0]);
+	JS::RootedString str(cx);
+	if (args.length() > 1 && (str = JS_ValueToString(cx, args[1])))
+		msg = JS_c_str(str);
+	if (debugLevel >= minlev && msg && *msg)
+		puts(msg);
+	cnzFree(msg);
+	args.rval().set(JSVAL_VOID);
+	return JS_TRUE;
+}				/* win_logputs */
+
 static JSBool win_prompt(JSContext * cx, unsigned int argc, jsval * vp)
 {
 	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
@@ -1867,6 +1937,7 @@ static JSBool win_confirm(JSContext * cx, unsigned int argc, jsval * vp)
 /* Set a timer or an interval */
 static JSObject *setTimeout(unsigned int argc, jsval * argv, bool isInterval)
 {
+	debugPrint(5, "set timer");
 	JS::RootedValue v0(jcx), v1(jcx);	// values of the 2 args
 // the function object, to execute, and the timer object.
 	JS::RootedObject fo(jcx, 0), to(jcx);
@@ -1878,12 +1949,15 @@ static JSObject *setTimeout(unsigned int argc, jsval * argv, bool isInterval)
 	const char *allocatedName = NULL;
 	const char *s = NULL;
 
-	if (argc != 2 || !JSVAL_IS_INT(argv[1]))
-		goto badarg;
-
 	v0 = argv[0];
-	v1 = argv[1];
-	n = JSVAL_TO_INT(v1);
+	if (argc != 2 || !JSVAL_IS_INT(argv[1])) {
+		// if only one parameter was supplied, hardcode
+		// a number of milliseconds.  1?  100?  1000?
+		n = 100;
+	} else {
+		v1 = argv[1];
+		n = JSVAL_TO_INT(v1);
+	}
 	if (JSVAL_IS_STRING(v0) ||
 	    v0.isObject() &&
 	    JS_ValueToObject(jcx, v0, fo.address()) &&
@@ -1893,7 +1967,7 @@ static JSObject *setTimeout(unsigned int argc, jsval * argv, bool isInterval)
 		to = JS_NewObject(jcx, &timer_class, NULL, winobj);
 		if (to == NULL) {
 abort:
-			misconfigure();
+			misconfigure(27);
 			return NULL;
 		}
 // protect this timer from the garbage collector
@@ -1936,6 +2010,7 @@ abort:
 				JS_ReportError(jcx,
 					       "error compiling function in %s()",
 					       methname);
+				debugPrint(5, "return");
 				return NULL;
 			}
 			strcpy(fname, "?");
@@ -1967,11 +2042,11 @@ abort:
 		effectChar((isInterval ? '1' : '0'));
 		endeffect();
 
+		debugPrint(5, "return");
 		return to;
 	}
 
-badarg:
-	JS_ReportError(jcx, "invalid arguments to %s()", methname);
+	debugPrint(5, "return");
 	return NULL;
 }				/* setTimeout */
 
@@ -2001,6 +2076,7 @@ static JSBool win_intv(JSContext * cx, unsigned int argc, jsval * vp)
 /* Clear a timer or an interval */
 static JSBool clearTimeout(JSContext * cx, unsigned int argc, jsval * vp)
 {
+	debugPrint(5, "clear timer");
 	JS::CallArgs args = JS::CallArgsFromVp(argc, vp);
 	args.rval().set(JSVAL_VOID);
 	if (argc == 0 || !args[0].isObject())
@@ -2010,11 +2086,13 @@ static JSBool clearTimeout(JSContext * cx, unsigned int argc, jsval * vp)
 	sprintf(nstring, "t{0|-|%s|0", pointer2string(obj));	// }
 	effectString(nstring);
 	endeffect();
+	debugPrint(5, "return");
 	return JS_TRUE;
 }				/* clearTimeout */
 
 static JSFunctionSpec window_methods[] = {
 	JS_FS("$puts$", win_puts, 1, 0),
+	JS_FS("$logputs$", win_logputs, 2, 0),
 	JS_FS("prompt", win_prompt, 2, 0),
 	JS_FS("confirm", win_confirm, 1, 0),
 	JS_FS("setTimeout", win_sto, 2, 0),
@@ -2199,7 +2277,7 @@ static void set_property_generic(js::HandleObject parent, const char *name)
 		propval = 0;
 		childroot = JS_NewObject(jcx, cp, NULL, parent);
 		if (!childroot) {
-			misconfigure();
+			misconfigure(28);
 			break;
 		}
 		if (cp == &url_class) {
@@ -2209,7 +2287,7 @@ static void set_property_generic(js::HandleObject parent, const char *name)
 			if (JS_DefineProperty
 			    (jcx, childroot, "href$val", v, NULL,
 			     setter_loc_hrefval, PROP_STD) == JS_FALSE) {
-				misconfigure();
+				misconfigure(29);
 				break;
 			}
 		}
@@ -2253,7 +2331,7 @@ static JSObject *instantiate_array1(js::HandleObject parent, const char *name)
 	v = OBJECT_TO_JSVAL(a);
 	if (JS_DefineProperty(jcx, parent, name, v, NULL, NULL, PROP_STD) ==
 	    JS_FALSE) {
-		misconfigure();
+		misconfigure(30);
 		return 0;
 	}
 	return a;
@@ -2285,7 +2363,7 @@ static JSObject *instantiate1(js::HandleObject parent, const char *name,
 	v = OBJECT_TO_JSVAL(a);
 	if (JS_DefineProperty(jcx, parent, name, v, NULL, NULL, PROP_STD) ==
 	    JS_FALSE) {
-		misconfigure();
+		misconfigure(31);
 		return 0;
 	}
 	return a;
@@ -2325,11 +2403,11 @@ set_array_element_object1(js::HandleObject parent, int idx,
 	JS_HasElement(jcx, parent, idx, &found);
 	if (found) {
 		if (JS_SetElement(jcx, parent, idx, v.address()) == JS_FALSE)
-			misconfigure();
+			misconfigure(32);
 	} else {
 		if (JS_DefineElement(jcx, parent, idx, v, NULL, NULL, PROP_STD)
 		    == JS_FALSE)
-			misconfigure();
+			misconfigure(33);
 	}
 }				/* set_array_element_object1 */
 
@@ -2439,18 +2517,31 @@ static void processMessage(void)
 	bool rc;		/* return code */
 	bool setret;		/* does setting a property produce a return? */
 	unsigned len;		/* array length */
+	char *uc_run;
+	int uc_len;
 
 	switch (head.cmd) {
 	case EJ_CMD_SCRIPT:
 		propval = 0;
 		s = runscript;
-/* Sometimes Mac puts these three chars at the start of a text file. */
+
+/* skip pass utf8 byte order mark if present */
 		if (!strncmp(s, "\xef\xbb\xbf", 3))
 			s += 3;
+
+/* Have to convert to unicode 16 to use JS_EvaluateUCScript.
+ * Manual says there is a version of EvaluateUCScript that converts the c string
+ * for us, but I can't find it anywhere, so I'm using utfHigh() in format.c,
+ * which I already wrote.
+ * This assumes little endian, last parameter. Is that always the case,
+ * perhaps a js standard, or does it follow the architecture of the machine? */
+		utfHigh(s, strlen(s), &uc_run, &uc_len, true, false, false);
+
 		head.n = 0;
 		head.proplength = 0;
-		if (JS_EvaluateScript(jcx, parent, s, strlen(s),
-				      "foo", head.lineno, v.address())) {
+		if (JS_EvaluateUCScript
+		    (jcx, parent, (const jschar *)uc_run, uc_len / 2,
+		     "foo", head.lineno, v.address())) {
 			if (v != JSVAL_VOID) {
 				s = 0;
 				JS::RootedString str(jcx);
@@ -2466,6 +2557,7 @@ static void processMessage(void)
 					head.proplength = strlen(s);
 			}
 		}
+		nzFree(uc_run);
 		nzFree(runscript);
 		runscript = 0;
 		writeHeader();
@@ -2560,7 +2652,7 @@ propreturn:
 			propval = 0;
 			child = JS_NewObject(jcx, cp, NULL, parent);
 			if (!child)
-				misconfigure();
+				misconfigure(34);
 			else
 				set_array_element_object1(parent, head.n,
 							  child);
