@@ -14,11 +14,13 @@
 
 const char *progname;
 const char eol[] = "\r\n";
-const char *version = "3.6.1";
+const char *version = "3.6.3";
 char *changeFileName;
 char *configFile, *addressFile, *cookieFile;
 char *mailDir, *mailUnread, *mailStash, *mailReply;
 char *recycleBin, *sigFile, *sigFileEnd;
+char *cacheDir;
+int cacheSize = 1000, cacheCount = 10000;
 char *ebTempDir, *ebUserDir;
 char *userAgents[10];
 char *currentAgent, *currentReferrer;
@@ -75,9 +77,6 @@ static pthread_mutex_t share_mutex = PTHREAD_MUTEX_INITIALIZER;
 static const char *javaDis[MAXNOJS];
 static int javaDisCount;
 
-#ifdef _MSC_VER
-#endif // _MSC_VER
-
 /*********************************************************************
 Redirect the incoming mail into a file, based on the subject or the sender.
 Along with the filters in .ebrc, this routine dips into your addressbook,
@@ -126,32 +125,22 @@ const char *mailRedirect(const char *to, const char *from,
 			break;
 
 		case 4:
-			if (stringEqualCI(m, subj))
-				return r;
-/* a prefix match is ok */
-			if (slen < 16 || mlen < 16)
-				break;	/* too short */
-			j = k = 0;
-			while (true) {
-				char c = subj[j];
-				char d = m[k];
-				if (isupperByte(c))
-					c = tolower(c);
-				if (isupperByte(d))
-					d = tolower(d);
-				if (!c || !d)
-					break;
-				if (c != d)
-					break;
-				for (++j; c == subj[j]; ++j) ;
-				for (++k; d == m[k]; ++k) ;
+			if (mlen > slen)
+				break;
+			if (mlen == slen) {
+				if (stringEqualCI(m, subj))
+					return r;
+				break;
 			}
-/* must match at least 2/3 of either string */
-			if (k > j)
-				j = k;
-			if (j >= 2 * mlen / 3 || j >= 2 * slen / 3) {
+/* a prefix or suffix match is ok */
+/* have to be at least half the subject line */
+			if (slen > mlen + mlen)
+				break;
+			if (memEqualCI(m, subj, mlen))
+				return r;
+			k = slen - mlen;
+			if (memEqualCI(m, subj + k, mlen))
 				return r;
-			}
 			break;
 		}		/* switch */
 	}			/* loop */
@@ -505,6 +494,7 @@ int main(int argc, char **argv)
 		}
 	}
 
+/* recycle bin and .signature files are unix-like, and not adjusted for windows. */
 	recycleBin = allocMem(strlen(home) + 8);
 	sprintf(recycleBin, "%s/.Trash", home);
 	if (fileTypeByName(recycleBin, false) != 'd') {
@@ -534,6 +524,8 @@ int main(int argc, char **argv)
 	strcat(agent0, version);
 	userAgents[0] = currentAgent = agent0;
 
+	setupEdbrowseTempDirectory();
+
 	progname = argv[0];
 	++argv, --argc;
 
@@ -553,8 +545,6 @@ int main(int argc, char **argv)
 			i_printfExit(MSG_Usage);
 	}
 
-	setupEdbrowseTempDirectory();
-
 	if (whichproc == 'j')
 		return js_main(argc, argv);
 
@@ -580,6 +570,7 @@ int main(int argc, char **argv)
 // with certificate file and cookie jar etc.
 	eb_curl_global_init();
 	cookiesFromJar();
+	setupEdbrowseCache();
 
 	for (; argc && argv[0][0] == '-'; ++argv, --argc) {
 		char *s = *argv;
@@ -722,8 +713,8 @@ int main(int argc, char **argv)
 		if (cx == 1)
 			runEbFunction("init");
 		changeFileName = 0;
-		cw->fileName = cloneString(file);
-		cw->firstURL = cloneString(file);
+		cf->fileName = cloneString(file);
+		cf->firstURL = cloneString(file);
 		if (isSQL(file))
 			cw->sqlMode = true;
 		rc = readFileArgv(file);
@@ -733,16 +724,16 @@ int main(int argc, char **argv)
 		if (!rc) {
 			showError();
 		} else if (changeFileName) {
-			nzFree(cw->fileName);
-			cw->fileName = changeFileName;
+			nzFree(cf->fileName);
+			cf->fileName = changeFileName;
 			changeFileName = 0;
 		}
 
 		cw->undoable = cw->changeMode = false;
 /* Browse the text if it's a url */
-		if (rc && isURL(cw->fileName) &&
-		    (cw->mt && cw->mt->outtype
-		     || isBrowseableURL(cw->fileName))) {
+		if (rc && isURL(cf->fileName) &&
+		    (cf->mt && cf->mt->outtype
+		     || isBrowseableURL(cf->fileName))) {
 			if (runCommand("b"))
 				debugPrint(1, "%d", fileSize);
 			else
@@ -1052,6 +1043,8 @@ void unreadConfigFile(void)
 	sslCerts = NULL;
 	downDir = NULL;
 	mailDir = NULL;
+	nzFree(cacheDir);
+	cacheDir = NULL;
 	nzFree(mailUnread);
 	mailUnread = NULL;
 	nzFree(mailReply);
@@ -1069,20 +1062,20 @@ void unreadConfigFile(void)
 /* Order is important here: mail{}, mime{}, table{}, then global keywords */
 #define MAILWORDS 0
 #define MIMEWORDS 8
-#define TABLEWORDS 15
-#define GLOBALWORDS 19
+#define TABLEWORDS 16
+#define GLOBALWORDS 20
 
 static const char *const keywords[] = {
 	"inserver", "outserver", "login", "password", "from", "reply",
 	"inport", "outport",
 	"type", "desc", "suffix", "protocol", "program",
-	"content", "outtype",
+	"content", "outtype", "urlmatch",
 	"tname", "tshort", "cols", "keycol",
-	"adbook", "downdir", "maildir", "agent",
-	"jar", "nojs", "xyz@xyz",
+	"downdir", "maildir", "agent",
+	"jar", "nojs", "cachedir",
 	"webtimer", "mailtimer", "certfile", "datasource", "proxy",
-	"linelength", "localizeweb", "jspool", "novs",
-	0
+	"linelength", "localizeweb", "jspool", "novs", "cachesize",
+	"adbook", 0
 };
 
 /* Read the config file and populate the corresponding data structures. */
@@ -1235,8 +1228,10 @@ bool readConfigFile(void)
 				q = last + 9;
 				if (*q == 0 || *q == '{' || *q == '(')
 					cfgLine0(MSG_EBRC_NoFnName);
+#if 0
 				if (isdigitByte(*q))
 					cfgLine0(MSG_EBRC_FnDigit);
+#endif
 				while (isalnumByte(*q))
 					++q;
 				if (q - last - 9 > 10)
@@ -1440,15 +1435,19 @@ putc:
 			mt->outtype = c;
 			continue;
 
-		case 15:	/* tname */
+		case 15:	/* urlmatch */
+			mt->urlmatch = v;
+			continue;
+
+		case 16:	/* tname */
 			td->name = v;
 			continue;
 
-		case 16:	/* tshort */
+		case 17:	/* tshort */
 			td->shortname = v;
 			continue;
 
-		case 17:	/* cols */
+		case 18:	/* cols */
 			while (*v) {
 				if (td->ncols == MAXTCOLS)
 					cfgLine1(MSG_EBRC_ManyCols, MAXTCOLS);
@@ -1461,7 +1460,7 @@ putc:
 			}
 			continue;
 
-		case 18:	/* keycol */
+		case 19:	/* keycol */
 			if (!isdigitByte(*v))
 				cfgLine0(MSG_EBRC_KeyNotNb);
 			td->key1 = strtol(v, &v, 10);
@@ -1471,13 +1470,6 @@ putc:
 				cfgLine1(MSG_EBRC_KeyOutRange, td->ncols);
 			continue;
 
-		case 19:	/* adbook */
-			addressFile = v;
-			ftype = fileTypeByName(v, false);
-			if (ftype && ftype != 'f')
-				cfgAbort1(MSG_EBRC_AbNotFile, v);
-			continue;
-
 		case 20:	/* downdir */
 			downDir = v;
 			if (fileTypeByName(v, false) != 'd')
@@ -1531,6 +1523,11 @@ putc:
 			javaDis[javaDisCount++] = v;
 			continue;
 
+		case 25:	/* cachedir */
+			nzFree(cacheDir);
+			cacheDir = cloneString(v);
+			continue;
+
 		case 26:	/* webtimer */
 			webTimeout = atoi(v);
 			continue;
@@ -1607,6 +1604,21 @@ putc:
 			addNovsHost(v);
 			continue;
 
+		case 35:	/* cachesize */
+			cacheSize = atoi(v);
+			if (cacheSize <= 0)
+				cacheSize = 0;
+			if (cacheSize >= 10000)
+				cacheSize = 10000;
+			continue;
+
+		case 36:	/* adbook */
+			addressFile = v;
+			ftype = fileTypeByName(v, false);
+			if (ftype && ftype != 'f')
+				cfgAbort1(MSG_EBRC_AbNotFile, v);
+			continue;
+
 		default:
 			cfgLine1(MSG_EBRC_KeywordNYI, s);
 		}		/* switch */
