@@ -313,7 +313,7 @@ enum {
 	MSG_ArrowAfter,
 	MSG_NoPrevious,
 	MSG_MAfter,
-	MSG_NoDestSession,
+	MSG_MovedSession,
 	MSG_NoBackup,
 	MSG_RangeG,
 	MSG_DBG,
@@ -464,7 +464,7 @@ enum {
 	MSG_MailAccountBadLineX,
 	MSG_TempNone,
 	MSG_MailFirstLine,
-	MSG_notused367,
+	MSG_ConvUtf16,
 	MSG_RecipNone3,
 	MSG_ProtExpected,
 	MSG_BadProt,
@@ -483,18 +483,18 @@ enum {
 	MSG_TempNoCreate2,
 	MSG_ManyTables,
 	MSG_RegularFile,
-	MSG_notused386,
+	MSG_ConvUtf32,
 	MSG_NoRead2,
 	MSG_NoAccess,
 	MSG_NoEnvVar,
 	MSG_NoTilde,
 	MSG_Submit,
-	MSG_notused392,
+	MSG_FrameNotHTML,
 	MSG_Secure,
 	MSG_Bymail,
 	MSG_Reset,
 	MSG_ShellExpand,
-	MSG_notused397,
+	MSG_FrameNoURL,
 	MSG_ShellManyMatch,
 	MSG_ShellLineLong,
 	MSG_DBUnspecified,
@@ -560,7 +560,7 @@ enum {
 	MSG_notused460,
 	MSG_AttAfterChars,
 	MSG_AttBad64,
-	MSG_notused463,
+	MSG_GlobalCommand2,
 	MSG_Doubleclick,
 	MSG_OptionSync,
 	MSG_XhrOff,
@@ -573,13 +573,13 @@ enum {
 	MSG_NJNoReset,
 	MSG_NJNoSubmit,
 	MSG_LostTag,
-	MSG_notused476,
+	MSG_ExpGo,
 	MSG_GarbledRefresh,
-	MSG_notused478,
+	MSG_NewWindow,
 	MSG_RedirectNoURL,
 	MSG_HTTPError,
-	MSG_notused481,
-	MSG_notused482,
+	MSG_NoFrame1,
+	MSG_NoFrame2,
 	MSG_HelpOn,
 	MSG_FTPDownload,
 	MSG_NoCertFile,
