@@ -298,8 +298,8 @@ empty:
 			}
 			sprintf(newfilename, "<buffer %d>", cx);
 			file = newfilename;
-			if (sessionList[cx].lw->fileName)
-				file = sessionList[cx].lw->fileName;
+			if (sessionList[cx].lw->f0.fileName)
+				file = sessionList[cx].lw->f0.fileName;
 		} else {
 			if (!fileIntoMemory(file, &buf, &buflen))
 				return false;
@@ -661,7 +661,7 @@ static char *buildSMTPURL(const struct MACCOUNT *account)
 	if (smlogin)
 		++smlogin;
 	else
-		smlogin = account->login;
+		smlogin = "unknown";
 
 	if (account->outssl & 1)
 		scheme = "smtps";
