@@ -16,7 +16,8 @@ The classes are created first, so that you can write meaningful prototypes here.
  * Better to have something than nothing at all. */
 height = 768;
 width = 1024;
-status = 0;
+// document.status is removed because it creates a conflict with
+// the status property of the XMLHttpRequest implementation
 defaultStatus = 0;
 returnValue = true;
 menubar = true;
@@ -30,6 +31,8 @@ document.bgcolor = "white";
 document.readyState = "loading";
 document.nodeType = 9;
 document.implementation = {};
+// pages seem to want document.style to exist
+document.style = {"bgcolor":"white"};
 
 screen = new Object;
 screen.height = 768;
@@ -64,7 +67,7 @@ return document.gebtn$(this, s);
 }
 document.gebtn$ = function(top, s) { 
 var a = new Array;
-if(s === '*' || (top.nodeName && top.nodeName === s))
+if(s === '*' || (top.nodeName && top.nodeName.toLowerCase() === s))
 a.push(top);
 if(top.childNodes) {
 for(var i=0; i<top.childNodes.length; ++i) {
@@ -81,7 +84,7 @@ return document.gebn$(this, s);
 }
 document.gebn$ = function(top, s) { 
 var a = new Array;
-if(s === '*' || (top.name && top.name === s))
+if(s === '*' || (top.name && top.name.toLowerCase() === s))
 a.push(top);
 if(top.childNodes) {
 for(var i=0; i<top.childNodes.length; ++i) {
@@ -98,7 +101,7 @@ return document.gebcn$(this, s);
 }
 document.gebcn$ = function(top, s) { 
 var a = new Array;
-if(s === '*' || (top.className && top.className === s))
+if(s === '*' || (top.className && top.className.toLowerCase() === s))
 a.push(top);
 if(top.childNodes) {
 for(var i=0; i<top.childNodes.length; ++i) {
@@ -190,6 +193,9 @@ break;
 case "div":
 c = new Div();
 break;
+case "p":
+c = new P();
+break;
 case "table":
 c = new Table();
 break;
@@ -217,8 +223,7 @@ c.value = "";
 return c;
 case "option":
 c = new Option();
-// don't need style or childNodes here.
-return c;
+break;
 default:
 /* $puts$("createElement default " + s); */
 c = new Element();
@@ -227,29 +232,43 @@ c = new Element();
 * but for most visible ones it does and I doubt it matters much */
 c.style = new Object;
 c.childNodes = new Array;
-c.attributes = new Object;
+c.attributes = new Array;
 c.nodeName = t;
+c.nodeType = 1;
+c.nodeValue = undefined;
+c.class = new String;
+c.className = new String;
+c.ownerDocument = document;
+c.tagName = t;
+
 return c;
 } 
 
 document.createTextNode = function(t) {
 var c = new TextNode(t);
 c.nodeName = "text";
+c.nodeValue = t;
+c.nodeType=3;
+c.ownerDocument = document;
+c.style = new Object;
+c.tagName = "text";
+c.className = new String;
 return c;
 }
 
 document.createDocumentFragment = function() {
 var c = document.createElement("fragment");
+c.nodeType = 11;
 return c;
 }
 
 document.createComment = function() {
 var c = document.createElement("comment");
+c.nodeType = 8;
 return c;
 }
 
 // window.alert is a simple wrapper around native puts.
-// Probably I should rename and wrapperize other functions like prompt and confirm.
 function alert(s) { $puts$(s); }
 
 /* window.open is the same as new window, just pass the args through */
@@ -457,6 +476,10 @@ return this.toString().lastIndexOf(s);
 URL.prototype.substring = function(from, to) { 
 return this.toString().substring(from, to);
 }
+// pages expect both substring and substr 
+URL.prototype.substr = function(from, to) {
+return this.toString().substr(from, to);
+}
 URL.prototype.toLowerCase = function() { 
 return this.toString().toLowerCase();
 }
@@ -469,6 +492,9 @@ return this.toString().match(s);
 URL.prototype.replace = function(s, t) { 
 return this.toString().replace(s, t);
 }
+URL.prototype.split = function(s) {
+return this.toString().split(s);
+}
 
 /*********************************************************************
 This is our addEventListener function.
@@ -497,7 +523,7 @@ this[ev] = undefined;
 }
 this[evarray] = a;
 eval(
-'this.' + ev + ' = function(){ var a = this.' + evarray + '; if(this.' + evorig + ') this.' + evorig + '(); for(var i = 0; i<a.length; ++i) a[i](); };');
+'this.' + ev + ' = function(){ var a = this.' + evarray + '; if(this.' + evorig + ') this.' + evorig + '(); for(var i = 0; i<a.length; ++i) {a[i]();} };');
 }
 this[evarray].push(handler);
 }
@@ -568,7 +594,30 @@ Anchor.prototype.addEventListener = window.addEventListener;
 Anchor.prototype.attachEvent = window.attachEvent;
 Anchor.prototype.removeEventListener = window.removeEventListener;
 
-/* document.appendChild and document.apch$ are native */
+/*********************************************************************
+The functions apch1$ and apch2$ are native. They perform appendChild in js.
+The first has no side effects, because the linkage was already performed
+within edbrowse via html, and a linkage side effect would only confuse things.
+The second, apch2$, has side effects, as js code calls appendChild
+and those links have to pass back to edbrowse.
+But, the actual function appendChild makes another check;
+if the child is already linked into the tree, then we have to unlink it first,
+before we put it somewhere else.
+This is a call to removeChild, which unlinks in js,
+and passses the remove side effect back to edbrowse.
+The same reasoning holds for insertBefore.
+*********************************************************************/
+
+document.appendChild = function(c) {
+if(c.parentNode) c.parentNode.removeChild(c);
+return this.apch2$(c);
+}
+
+document.insertBefore = function(c, t) {
+if(c.parentNode) c.parentNode.removeChild(c);
+return this.insbf$(c, t);
+}
+
 document.childNodes = new Array;
 Object.defineProperty(document, "firstChild", {
 get: function() { return document.childNodes[0]; }
@@ -576,6 +625,13 @@ get: function() { return document.childNodes[0]; }
 Object.defineProperty(document, "lastChild", {
 get: function() { return document.childNodes[document.childNodes.length-1]; }
 });
+Object.defineProperty(document, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(document, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 document.hasChildNodes = function() { return (this.childNodes.length > 0); }
 document.replaceChild = function(newc, oldc) {
 var lastentry;
@@ -599,9 +655,27 @@ break;
 }
 }
 document.getAttribute = function(name) { return this[name.toLowerCase()]; }
+document.hasAttribute = function(name) { if (this[name.toLowerCase()]) return true; else return false; }
+// Sets the attribute in 3 different ways, the third using attributes as an array.
+// This may be overkill. I don't know.
 document.setAttribute = function(name, v) { 
-this.attributes[name.toLowerCase()] = v;
-this[name.toLowerCase()] = v; 
+var n = name.toLowerCase();
+this[n] = v; 
+this.attributes[n] = v;
+this.attributes.push(n);
+}
+// Removes the attribute in 3 different ways, the third using attributes as an array.
+// This may be overkill. I don't know.
+document.removeAttribute = function(name) {
+    var n = name.toLowerCase();
+    if (this[n]) delete this[n];
+if(this.attributes[n]) delete this.attributes[n];
+    for (var i=this.attributes.length - 1; i >= 0; --i) {
+        if (this.attributes[i] == n) {
+this.attributes.splice(i, 1);
+break;
+}
+}
 }
 
 /*********************************************************************
@@ -685,6 +759,35 @@ nodeToReturn[url] = new URL(u.href);
 return nodeToReturn;
 }
 
+function get_sibling(obj,direction)
+{
+if (typeof obj.parentNode == 'undefined') {
+// need calling node to have parent and it doesn't, error
+return null;
+}
+var pn = obj.parentNode;
+var j, l;
+l = pn.childNodes.length;
+for (j=0; j<l; ++j)
+if (pn.childNodes[j] == obj) break;
+if (j == l) {
+// child not found under parent, error
+return null;
+}
+switch(direction)
+{
+case "previous":
+return (j > 0 ? pn.childNodes[j-1] : null);
+break;
+case "next":
+return (j < l-1 ? pn.childNodes[j+1] : null);
+break;
+default:
+// the function should always have been called with either 'previous' or 'next' specified
+return null;
+}
+}
+
 /* The select element in a form is itself an array, so the above functions have
  * to be on array prototype, except appendchild is to have no side effects,
  * because select options are maintained by rebuildSelectors(), so appendChild
@@ -721,49 +824,80 @@ get: function() { return this[0]; }
 Object.defineProperty(Array.prototype, "lastChild", {
 get: function() { return this[this.length-1]; }
 });
+Object.defineProperty(Array.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Array.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Array.prototype.hasChildNodes = document.hasChildNodes;
 Array.prototype.replaceChild = document.replaceChild;
 Array.prototype.getAttribute = document.getAttribute;
 Array.prototype.setAttribute = document.setAttribute;
+Array.prototype.hasAttribute = document.hasAttribute;
+Array.prototype.removeAttribute = document.removeAttribute;
 
 Head.prototype.appendChild = document.appendChild;
-Head.prototype.apch$ = document.apch$;
+Head.prototype.apch1$ = document.apch1$;
+Head.prototype.apch2$ = document.apch2$;
 Head.prototype.insertBefore = document.insertBefore;
+Head.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Head.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Head.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Head.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Head.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Head.prototype.hasChildNodes = document.hasChildNodes;
 Head.prototype.removeChild = document.removeChild;
 Head.prototype.replaceChild = document.replaceChild;
 Head.prototype.getAttribute = document.getAttribute;
 Head.prototype.setAttribute = document.setAttribute;
 Head.prototype.cloneNode = document.cloneNode;
+Head.prototype.hasAttribute = document.hasAttribute;
+Head.prototype.removeAttribute = document.removeAttribute;
 
 Body.prototype.appendChild = document.appendChild;
-Body.prototype.apch$ = document.apch$;
+Body.prototype.apch1$ = document.apch1$;
+Body.prototype.apch2$ = document.apch2$;
 Body.prototype.insertBefore = document.insertBefore;
+Body.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Body.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Body.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Body.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Body.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Body.prototype.hasChildNodes = document.hasChildNodes;
 Body.prototype.removeChild = document.removeChild;
 Body.prototype.replaceChild = document.replaceChild;
 Body.prototype.getAttribute = document.getAttribute;
 Body.prototype.setAttribute = document.setAttribute;
 Body.prototype.cloneNode = document.cloneNode;
-
+Body.prototype.hasAttribute = document.hasAttribute;
+Body.prototype.removeAttribute = document.removeAttribute;
 
 /*********************************************************************
 Special functions for form and input.
 If you add an input to a form, it adds under childNodes in the usual way,
 but also must add in the elements[] array.
 Same for insertBefore and removeChild.
+Same for insertBefore and removeChild.
 When adding an input element to a form,
 linnk form[element.name] to that element.
 *********************************************************************/
@@ -789,7 +923,9 @@ this.elements.appendChild(newobj);
 form$name(this, newobj);
 }
 }
-Form.prototype.apch$ = document.apch$;
+Form.prototype.apch1$ = document.apch1$;
+Form.prototype.apch2$ = document.apch2$;
+Form.prototype.insbf$ = document.insbf$;
 Form.prototype.insertBeforeNative = document.insertBefore;
 Form.prototype.insertBefore = function(newobj, item) {
 this.insertBeforeNative(newobj, item);
@@ -805,6 +941,13 @@ get: function() { return this.childNodes[0]; }
 Object.defineProperty(Form.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Form.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Form.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Form.prototype.hasChildNodes = document.hasChildNodes;
 Form.prototype.removeChildNative = document.removeChild;
 Form.prototype.removeChild = function(item) {
@@ -816,16 +959,27 @@ Form.prototype.replaceChild = document.replaceChild;
 Form.prototype.getAttribute = document.getAttribute;
 Form.prototype.setAttribute = document.setAttribute;
 Form.prototype.cloneNode = document.cloneNode;
+Form.prototype.hasAttribute = document.hasAttribute;
+Form.prototype.removeAttribute = document.removeAttribute;
 
 Element.prototype.appendChild = document.appendChild;
-Element.prototype.apch$ = document.apch$;
+Element.prototype.apch1$ = document.apch1$;
+Element.prototype.apch2$ = document.apch2$;
 Element.prototype.insertBefore = document.insertBefore;
+Element.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Element.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Element.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Element.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Element.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Element.prototype.hasChildNodes = document.hasChildNodes;
 Element.prototype.removeChild = document.removeChild;
 Element.prototype.replaceChild = document.replaceChild;
@@ -834,179 +988,427 @@ Element.prototype.setAttribute = document.setAttribute;
 Element.prototype.focus = document.focus;
 Element.prototype.blur = document.blur;
 Element.prototype.cloneNode = document.cloneNode;
+Element.prototype.hasAttribute = document.hasAttribute;
+Element.prototype.removeAttribute = document.removeAttribute;
 
 Anchor.prototype.appendChild = document.appendChild;
-Anchor.prototype.apch$ = document.apch$;
+Anchor.prototype.apch1$ = document.apch1$;
+Anchor.prototype.apch2$ = document.apch2$;
 Anchor.prototype.focus = document.focus;
 Anchor.prototype.blur = document.blur;
 Anchor.prototype.getAttribute = document.getAttribute;
 Anchor.prototype.setAttribute = document.setAttribute;
 Anchor.prototype.cloneNode = document.cloneNode;
+Anchor.prototype.hasAttribute = document.hasAttribute;
+Anchor.prototype.removeAttribute = document.removeAttribute;
 
 Div.prototype.appendChild = document.appendChild;
-Div.prototype.apch$ = document.apch$;
+Div.prototype.apch1$ = document.apch1$;
+Div.prototype.apch2$ = document.apch2$;
 Div.prototype.insertBefore = document.insertBefore;
+Div.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Div.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Div.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Div.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Div.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Div.prototype.hasChildNodes = document.hasChildNodes;
 Div.prototype.removeChild = document.removeChild;
 Div.prototype.replaceChild = document.replaceChild;
 Div.prototype.getAttribute = document.getAttribute;
 Div.prototype.setAttribute = document.setAttribute;
 Div.prototype.cloneNode = document.cloneNode;
+Div.prototype.hasAttribute = document.hasAttribute;
+Div.prototype.removeAttribute = document.removeAttribute;
 
 HtmlObj.prototype.appendChild = document.appendChild;
-HtmlObj.prototype.apch$ = document.apch$;
+HtmlObj.prototype.apch1$ = document.apch1$;
+HtmlObj.prototype.apch2$ = document.apch2$;
 HtmlObj.prototype.insertBefore = document.insertBefore;
+HtmlObj.prototype.insbf$ = document.insbf$;
 Object.defineProperty(HtmlObj.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(HtmlObj.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(HtmlObj.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(HtmlObj.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
+
 HtmlObj.prototype.hasChildNodes = document.hasChildNodes;
 HtmlObj.prototype.removeChild = document.removeChild;
 HtmlObj.prototype.replaceChild = document.replaceChild;
 HtmlObj.prototype.getAttribute = document.getAttribute;
 HtmlObj.prototype.setAttribute = document.setAttribute;
 HtmlObj.prototype.cloneNode = document.cloneNode;
+HtmlObj.prototype.hasAttribute = document.hasAttribute;
+HtmlObj.prototype.removeAttribute = document.removeAttribute;
 
 Script.prototype.getAttribute = document.getAttribute;
 Script.prototype.setAttribute = document.setAttribute;
 Script.prototype.cloneNode = document.cloneNode;
+Script.prototype.hasAttribute = document.hasAttribute;
+Script.prototype.removeAttribute = document.removeAttribute;
+Script.prototype.appendChild = document.appendChild;
+Script.prototype.apch1$ = document.apch1$;
+Script.prototype.apch2$ = document.apch2$;
+Script.prototype.insertBefore = document.insertBefore;
+Script.prototype.insbf$ = document.insbf$;
+Script.prototype.hasChildNodes = document.hasChildNodes;
+Script.prototype.removeChild = document.removeChild;
+Script.prototype.replaceChild = document.replaceChild;
+
+Object.defineProperty(Script.prototype, "firstChild", {
+get: function() { return this.childNodes[0]; }
+});
+Object.defineProperty(Script.prototype, "lastChild", {
+get: function() { return this.childNodes[this.childNodes.length-1]; }
+});
+Object.defineProperty(Script.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Script.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
+TextNode.prototype.getAttribute = document.getAttribute;
+TextNode.prototype.setAttribute = document.setAttribute;
+TextNode.prototype.cloneNode = document.cloneNode;
+TextNode.prototype.hasAttribute = document.hasAttribute;
+TextNode.prototype.removeAttribute = document.removeAttribute;
+TextNode.prototype.appendChild = document.appendChild;
+TextNode.prototype.apch1$ = document.apch1$;
+TextNode.prototype.apch2$ = document.apch2$;
+TextNode.prototype.insertBefore = document.insertBefore;
+TextNode.prototype.insbf$ = document.insbf$;
+TextNode.prototype.hasChildNodes = document.hasChildNodes;
+TextNode.prototype.removeChild = document.removeChild;
+TextNode.prototype.replaceChild = document.replaceChild;
+
+Object.defineProperty(TextNode.prototype, "firstChild", {
+get: function() { return this.childNodes[0]; }
+});
+Object.defineProperty(TextNode.prototype, "lastChild", {
+get: function() { return this.childNodes[this.childNodes.length-1]; }
+});
+Object.defineProperty(TextNode.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(TextNode.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
 
 P.prototype.appendChild = document.appendChild;
-P.prototype.apch$ = document.apch$;
+P.prototype.apch1$ = document.apch1$;
+P.prototype.apch2$ = document.apch2$;
 P.prototype.getAttribute = document.getAttribute;
 P.prototype.setAttribute = document.setAttribute;
+P.prototype.hasAttribute = document.hasAttribute;
+P.prototype.removeAttribute = document.removeAttribute;
+
 P.prototype.insertBefore = document.insertBefore;
+P.prototype.insbf$ = document.insbf$;
 Object.defineProperty(P.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(P.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(P.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(P.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 P.prototype.hasChildNodes = document.hasChildNodes;
 P.prototype.removeChild = document.removeChild;
 P.prototype.replaceChild = document.replaceChild;
 P.prototype.cloneNode = document.cloneNode;
 
 Lister.prototype.appendChild = document.appendChild;
-Lister.prototype.apch$ = document.apch$;
+Lister.prototype.apch1$ = document.apch1$;
+Lister.prototype.apch2$ = document.apch2$;
 Lister.prototype.getAttribute = document.getAttribute;
 Lister.prototype.setAttribute = document.setAttribute;
 Lister.prototype.insertBefore = document.insertBefore;
+Lister.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Lister.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Lister.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Lister.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Lister.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Lister.prototype.hasChildNodes = document.hasChildNodes;
 Lister.prototype.removeChild = document.removeChild;
 Lister.prototype.replaceChild = document.replaceChild;
 Lister.prototype.cloneNode = document.cloneNode;
+Lister.prototype.hasAttribute = document.hasAttribute;
+Lister.prototype.removeAttribute = document.removeAttribute;
 
 Listitem.prototype.appendChild = document.appendChild;
-Listitem.prototype.apch$ = document.apch$;
+Listitem.prototype.apch1$ = document.apch1$;
+Listitem.prototype.apch2$ = document.apch2$;
 Listitem.prototype.getAttribute = document.getAttribute;
 Listitem.prototype.setAttribute = document.setAttribute;
 Listitem.prototype.insertBefore = document.insertBefore;
+Listitem.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Listitem.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Listitem.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Listitem.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Listitem.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Listitem.prototype.hasChildNodes = document.hasChildNodes;
 Listitem.prototype.removeChild = document.removeChild;
 Listitem.prototype.replaceChild = document.replaceChild;
 Listitem.prototype.cloneNode = document.cloneNode;
+Listitem.prototype.hasAttribute = document.hasAttribute;
+Listitem.prototype.removeAttribute = document.removeAttribute;
 
 Table.prototype.appendChild = document.appendChild;
-Table.prototype.apch$ = document.apch$;
+Table.prototype.apch1$ = document.apch1$;
+Table.prototype.apch2$ = document.apch2$;
 Table.prototype.getAttribute = document.getAttribute;
 Table.prototype.setAttribute = document.setAttribute;
 Table.prototype.insertBefore = document.insertBefore;
+Table.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Table.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Table.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Table.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Table.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Table.prototype.hasChildNodes = document.hasChildNodes;
 Table.prototype.removeChild = document.removeChild;
 Table.prototype.replaceChild = document.replaceChild;
 Table.prototype.cloneNode = document.cloneNode;
+Table.prototype.hasAttribute = document.hasAttribute;
+Table.prototype.removeAttribute = document.removeAttribute;
 
 Tbody.prototype.appendChild = document.appendChild;
-Tbody.prototype.apch$ = document.apch$;
+Tbody.prototype.apch1$ = document.apch1$;
+Tbody.prototype.apch2$ = document.apch2$;
 Tbody.prototype.getAttribute = document.getAttribute;
 Tbody.prototype.setAttribute = document.setAttribute;
 Tbody.prototype.insertBefore = document.insertBefore;
+Tbody.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Tbody.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Tbody.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Tbody.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Tbody.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Tbody.prototype.hasChildNodes = document.hasChildNodes;
 Tbody.prototype.removeChild = document.removeChild;
 Tbody.prototype.replaceChild = document.replaceChild;
 Tbody.prototype.cloneNode = document.cloneNode;
+Tbody.prototype.hasAttribute = document.hasAttribute;
+Tbody.prototype.removeAttribute = document.removeAttribute;
 
 Trow.prototype.appendChild = document.appendChild;
-Trow.prototype.apch$ = document.apch$;
+Trow.prototype.apch1$ = document.apch1$;
+Trow.prototype.apch2$ = document.apch2$;
 Trow.prototype.getAttribute = document.getAttribute;
 Trow.prototype.setAttribute = document.setAttribute;
 Trow.prototype.insertBefore = document.insertBefore;
+Trow.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Trow.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Trow.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Trow.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Trow.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Trow.prototype.hasChildNodes = document.hasChildNodes;
 Trow.prototype.removeChild = document.removeChild;
 Trow.prototype.replaceChild = document.replaceChild;
 Trow.prototype.cloneNode = document.cloneNode;
+Trow.prototype.hasAttribute = document.hasAttribute;
+Trow.prototype.removeAttribute = document.removeAttribute;
 
 Cell.prototype.appendChild = document.appendChild;
-Cell.prototype.apch$ = document.apch$;
+Cell.prototype.apch1$ = document.apch1$;
+Cell.prototype.apch2$ = document.apch2$;
 Cell.prototype.getAttribute = document.getAttribute;
 Cell.prototype.setAttribute = document.setAttribute;
 Cell.prototype.insertBefore = document.insertBefore;
+Cell.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Cell.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Cell.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Cell.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Cell.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
+
 Cell.prototype.hasChildNodes = document.hasChildNodes;
 Cell.prototype.removeChild = document.removeChild;
 Cell.prototype.replaceChild = document.replaceChild;
 Cell.prototype.cloneNode = document.cloneNode;
+Cell.prototype.hasAttribute = document.hasAttribute;
+Cell.prototype.removeAttribute = document.removeAttribute;
 
 Span.prototype.appendChild = document.appendChild;
-Span.prototype.apch$ = document.apch$;
+Span.prototype.apch1$ = document.apch1$;
+Span.prototype.apch2$ = document.apch2$;
 Span.prototype.getAttribute = document.getAttribute;
 Span.prototype.setAttribute = document.setAttribute;
 Span.prototype.insertBefore = document.insertBefore;
+Span.prototype.insbf$ = document.insbf$;
 Object.defineProperty(Span.prototype, "firstChild", {
 get: function() { return this.childNodes[0]; }
 });
 Object.defineProperty(Span.prototype, "lastChild", {
 get: function() { return this.childNodes[this.childNodes.length-1]; }
 });
+Object.defineProperty(Span.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Span.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
 Span.prototype.hasChildNodes = document.hasChildNodes;
 Span.prototype.removeChild = document.removeChild;
 Span.prototype.replaceChild = document.replaceChild;
 Span.prototype.cloneNode = document.cloneNode;
+Span.prototype.hasAttribute = document.hasAttribute;
+Span.prototype.removeAttribute = document.removeAttribute;
 
+Image.prototype.appendChild = document.appendChild;
+Image.prototype.apch1$ = document.apch1$;
+Image.prototype.apch2$ = document.apch2$;
+Image.prototype.getAttribute = document.getAttribute;
+Image.prototype.setAttribute = document.setAttribute;
+Image.prototype.insertBefore = document.insertBefore;
+Image.prototype.insbf$ = document.insbf$;
+Object.defineProperty(Image.prototype, "firstChild", {
+get: function() { return this.childNodes[0]; }
+});
+Object.defineProperty(Image.prototype, "lastChild", {
+get: function() { return this.childNodes[this.childNodes.length-1]; }
+});
+Object.defineProperty(Image.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Image.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
+Image.prototype.hasChildNodes = document.hasChildNodes;
+Image.prototype.removeChild = document.removeChild;
+Image.prototype.replaceChild = document.replaceChild;
+Image.prototype.cloneNode = document.cloneNode;
+Image.prototype.hasAttribute = document.hasAttribute;
+Image.prototype.removeAttribute = document.removeAttribute;
+
+Frame.prototype.appendChild = document.appendChild;
+Frame.prototype.apch1$ = document.apch1$;
+Frame.prototype.apch2$ = document.apch2$;
+Frame.prototype.getAttribute = document.getAttribute;
+Frame.prototype.setAttribute = document.setAttribute;
+Frame.prototype.insertBefore = document.insertBefore;
+Frame.prototype.insbf$ = document.insbf$;
+Object.defineProperty(Frame.prototype, "firstChild", {
+get: function() { return this.childNodes[0]; }
+});
+Object.defineProperty(Frame.prototype, "lastChild", {
+get: function() { return this.childNodes[this.childNodes.length-1]; }
+});
+Object.defineProperty(Frame.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Frame.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
+Frame.prototype.hasChildNodes = document.hasChildNodes;
+Frame.prototype.removeChild = document.removeChild;
+Frame.prototype.replaceChild = document.replaceChild;
+Frame.prototype.cloneNode = document.cloneNode;
+Frame.prototype.hasAttribute = document.hasAttribute;
+Frame.prototype.removeAttribute = document.removeAttribute;
+
+Option.prototype.appendChild = document.appendChild;
+Option.prototype.apch1$ = document.apch1$;
+Option.prototype.apch2$ = document.apch2$;
+Option.prototype.getAttribute = document.getAttribute;
+Option.prototype.setAttribute = document.setAttribute;
+Option.prototype.insertBefore = document.insertBefore;
+Option.prototype.insbf$ = document.insbf$;
+Object.defineProperty(Option.prototype, "firstChild", {
+get: function() { return this.childNodes[0]; }
+});
+Object.defineProperty(Option.prototype, "lastChild", {
+get: function() { return this.childNodes[this.childNodes.length-1]; }
+});
+Object.defineProperty(Option.prototype, "nextSibling", {
+get: function() { return get_sibling(this,"next"); }
+});
+Object.defineProperty(Option.prototype, "previousSibling", {
+get: function() { return get_sibling(this,"previous"); }
+});
+
+Option.prototype.hasChildNodes = document.hasChildNodes;
+Option.prototype.removeChild = document.removeChild;
+Option.prototype.replaceChild = document.replaceChild;
+Option.prototype.cloneNode = document.cloneNode;
+Option.prototype.hasAttribute = document.hasAttribute;
+Option.prototype.removeAttribute = document.removeAttribute;
 
 /* navigator; some parameters are filled in by the buildstartwindow script. */
 navigator.appName = "edbrowse";
@@ -1033,9 +1435,8 @@ history.toString = function() {
  return "Sorry, edbrowse does not maintain a browsing history.";
 } 
 
-/* The web console, just the log method for now, and only one arg. */
-console = new Object;
-console.log = function(obj) {
+/* The web console, one argument, print based on debugLevel */
+time$log$ = function(debug, level, obj) {
 var today=new Date();
 var h=today.getHours();
 var m=today.getMinutes();
@@ -1044,11 +1445,13 @@ var s=today.getSeconds();
 if(h < 10) h = "0" + h;
 if(m < 10) m = "0" + m;
 if(s < 10) s = "0" + s;
-$puts$("[" + h + ":" + m + ":" + s + "] " + obj);
+$logputs$(debug, "console " + level + " [" + h + ":" + m + ":" + s + "] " + obj);
 }
-console.info = console.log;
-console.warn = console.log;
-console.error = console.log;
+console = new Object;
+console.log = function(obj) { time$log$(3, "log", obj); }
+console.info = function(obj) { time$log$(3, "info", obj); }
+console.warn = function(obj) { time$log$(3, "warn", obj); }
+console.error = function(obj) { time$log$(3, "error", obj); }
 
 /* An ok (object keys) function for javascript/dom debugging. */
 /* This is in concert with the jdb command in edbrowse. */
@@ -1108,6 +1511,8 @@ this.async = false;
 
 this.method = method || "GET";
 this.url = convert_url(url);
+this.status = 0;
+this.statusText = "";
 this.onreadystatechange();
 },
 setRequestHeader: function(header, value){
@@ -1152,6 +1557,8 @@ this.readyState = 4;
 
 if ((!this.aborted) && this.responseText.length > 0){
 this.readyState = 4;
+this.status = 200;
+this.statusText = "OK";
 this.onreadystatechange();
 }
 
@@ -1221,8 +1628,55 @@ if (protocol1 == sideprotocol)
 new_url = path;
 // ok - it is fully qualified already
 } else {
+
+sidehost_last = sidehost.substring(sidehost.length-1,sidehost.length);
+path_first = path.substring(0,1);
+if (sidehost_last !== '/' && path_first !== '/')
+{
+new_url = sideprotocol + '//' + sidehost + '/' + path;
+} else {
 new_url = sideprotocol + '//' + sidehost + path;
 }
+}
 return new_url;
 }
 
+document.defaultView = function()
+{
+return this.style;
+}
+
+document.defaultView.getComputedStyle = function()
+{
+        obj = new CSSStyleDeclaration;
+        obj.element = document;
+        obj.style = document.style;
+        return obj;
+}
+
+getComputedStyle = function(n)
+{
+        obj = new CSSStyleDeclaration;
+        obj.element = this;
+        obj.style = new Array;
+        obj.style.push({n:obj.style[n]});
+        return obj;
+}
+
+CSSStyleDeclaration = function(){
+        this.element = null;
+        this.style = null;
+};
+
+CSSStyleDeclaration.prototype = {
+getPropertyValue: function (n)
+        {
+                if (this.style[n] == undefined)
+                {
+                        this.style[n] = 0;
+                        return 0;
+                } else {
+                        return this.style[n];
+                }
+        }
+}
