@@ -232,6 +232,10 @@ static bool httpDefault(const char *url)
 		if (s == end)
 			end = colon;
 	}
+// only domain characters allowed
+	for (s = url; s < end; ++s)
+		if (!isalnumByte(*s) && *s != '.' && *s != '-')
+			return false;
 /* need at least two embedded dots */
 	n = 0;
 	for (s = url + 1; s < end - 1; ++s)
@@ -255,7 +259,7 @@ static bool httpDefault(const char *url)
 		    && !domainSuffix[n][len])
 			return true;
 /* www.anything.xx is ok */
-	if (len == 2 && memEqualCI(url, "www.", 4))
+	if (len >= 2 && memEqualCI(url, "www.", 4))
 		return true;
 	return false;
 }				/* httpDefault */
@@ -272,7 +276,7 @@ static int parseURL(const char *url, const char **proto, int *prlen, const char
 		    const char **portloc, int *port,
 		    const char **data, int *dalen, const char **post)
 {
-	const char *p, *q;
+	const char *p, *q, *pp;
 	int a;
 
 	if (proto)
@@ -377,7 +381,7 @@ static int parseURL(const char *url, const char **proto, int *prlen, const char
 /* find the end of the domain */
 	q = p + strcspn(p, "@?#/\1");
 	if (*q == '@') {	/* user:password@host */
-		const char *pp = strchr(p, ':');
+		pp = strchr(p, ':');
 		if (!pp || pp > q) {	/* no password */
 			if (user)
 				*user = p;
@@ -398,6 +402,10 @@ static int parseURL(const char *url, const char **proto, int *prlen, const char
 
 /* again, look for the end of the domain */
 	q = p + strcspn(p, ":?#/\1");
+// only domain characters allowed
+	for (pp = p; pp < q; ++pp)
+		if (!isalnumByte(*pp) && *pp != '.' && *pp != '-')
+			return -1;
 	if (host)
 		*host = p;
 	if (holen)
@@ -802,7 +810,9 @@ char *resolveURL(const char *base, const char *rel)
 	n = allocString(strlen(base) + strlen(rel) + 12);
 	debugPrint(5, "resolve(%s|%s)", base, rel);
 
+#if 0
 // This is specifically to work around a tidy bug.
+// fixed in 5.1.36 so not needed any more.
 	for (s = rel; *s == '%'; s += 3) {
 		if (s[1] == '2' && s[2] == '0')
 			continue;
@@ -811,6 +821,7 @@ char *resolveURL(const char *base, const char *rel)
 	if (s > rel && memEqualCI(s, "http", 4))
 		rel = s;
 // end tidy workaround
+#endif
 
 	if (rel[0] == '#' && !strchr(rel, '/')) {
 /* This is an anchor for the current document, don't resolve. */
@@ -942,14 +953,13 @@ bool sameURL(const char *s, const char *t)
 	return !memcmp(s, t, l);
 }				/* sameURL */
 
-/* Find some helpful text to print, in place of an image.
+/* Find some helpful text to print in place of an image.
  * Not sure why we would need more than 1000 chars for this,
  * so return a static buffer. */
 char *altText(const char *base)
 {
 	static char buf[1000];
 	int len, n;
-	int recount = 0;
 	char *s;
 	debugPrint(6, "altText(%s)", base);
 	if (!base)
@@ -958,24 +968,18 @@ char *altText(const char *base)
 		return 0;
 	if (memEqualCI(base, "javascript", 10))
 		return 0;
-
-retry:
-	if (recount >= 2)
-		return 0;
 	strncpy(buf, base, sizeof(buf) - 1);
 	spaceCrunch(buf, true, false);
 	len = strlen(buf);
-	if (len && !isalnumByte(buf[len - 1]))
+/* remove punctuation mark from the end of a sentence or phrase */
+	if (len >= 2 && !isalnumByte(buf[len - 1]) && isalnumByte(buf[len - 2]))
 		buf[--len] = 0;
-	while (len && !isalnumByte(buf[0]))
+/* strip leading whitespace */
+	while (len && isspaceByte(buf[0]))
 		strmove(buf, buf + 1), --len;
 	if (len > 10) {
 /* see whether it's a phrase/sentence or a pathname/url */
-/* Do this by counting spaces */
-		for (n = 0, s = buf; *s; ++s)
-			if (*s == ' ')
-				++n;
-		if (n * 8 >= len)
+		if (!isURL(buf))
 			return buf;	/* looks like words */
 /* Ok, now we believe it's a pathname or url */
 /* get rid of post or get data */
@@ -998,31 +1002,16 @@ retry:
 				*s = 0;
 		}
 /* Get rid of everything up to the last slash, leaving the file name */
+retry:
 		s = strrchr(buf, '/');
-		if (s && recount) {
-			char *ss;
-			*s = 0;
-			ss = strrchr(buf, '/');
-			if (!ss)
-				return 0;
-			if (ss > buf && ss[-1] == '/')
-				return 0;
-			*s = '/';
-			s = ss;
-		}
-		if (s)
+		if (s && s - buf >= 12) {
+			if (!s[1]) {
+				*s = 0;
+				goto retry;
+			}
 			strmove(buf, s + 1);
+		}
 	}			/* more than ten characters */
-	++recount;
-/* If we don't have enough letters, forget it */
-	len = strlen(buf);
-	if (len < 3)
-		goto retry;
-	for (n = 0, s = buf; *s; ++s)
-		if (isalphaByte(*s))
-			++n;
-	if (n * 2 <= len)
-		return 0;	/* not enough letters */
 	return buf;
 }				/* altText */
 
